/* clmb.c - Cluster Member test program */

/* Written 1999 by Mohsen.Souissi@inria.fr, INRIA */

#include <sys/types.h>
#include <sys/socket.h>
#include <net/if.h>
#include <netinet/in.h>
#include <netnatm/natm.h>
#include <arpa/inet.h>
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <signal.h>
#include <limits.h>
#include <sys/wait.h>
#include <sys/queue.h>

#include "../linux-atm.h"
#include "atm.h"
#include "atmd.h"
#include "ipv6_tools.h"

#include "atmcall.h"
#include "scl_io.h"

#include "tables.h"
#include "mars_msg.h"
#include "mars_def.h"

#include "mcl_tab.h"
#include "mcl_io.h"

#define COMPONENT	"MAIN"

int debug = 0;
int myitf = 0;

const char *sig_qos = NULL;
const char *dump_dir = NULL;



/* in case of failiure to send request to MARS... see after... @@@ */
int reopen_mars(void)
{
    return(0);
}

static int main_loop(void)
{
    int error = 0, fin = 0, fds;
    fd_set set, perm;
    ATMCALL *call;
    pid_t pid;
    int sig_down = 0, sd, rt_sd;
    enum atmsvc_msg_type type;
    
    if (signal(SIGPIPE, sig_pipe) == SIG_ERR)
	diag(COMPONENT, DIAG_ERROR, "main_loop: sig_pipe");
  
    FD_ZERO(&perm);
    FD_SET(sig_sd, &perm);
    fds = FD_SETSIZE;
    (void) gettimeofday(&now, NULL);

    if ((error = socketpair(AF_UNIX, SOCK_STREAM, 0, psd)) < 0) {
	diag(COMPONENT, DIAG_ERROR, 
	     "main_loop: failed to create a socketpair");
	return(-1);
    }

    /* create a socket for routing messages */
    if ((rt_sd = socket(PF_ROUTE, SOCK_RAW, 0)) < 0) {
	perror("socket");
	diag(COMPONENT, DIAG_ERROR, 
	     "main_loop: failed to create a routing socket");
	return(-1);
    }

    if ((pid = fork()) < 0) {
	diag(COMPONENT, DIAG_ERROR, "fork error: %s", strerror(errno));
	exit(1);
    }

    if (pid == 0) {		/* child */
	(void) close(sig_sd);	/* child doesn't use it */
	(void) close(rt_sd);
	if ((fin = user_input()) == 1) {
	    diag(COMPONENT, DIAG_DEBUG, "Child exited normally");
	    exit(0);
	}
	exit(1);
    }
    if (pid > 0) {
	while (fin <= 0) {
	    set = perm;
	    FD_SET(psd[0], &set);
	    FD_SET(rt_sd, &set);
	    

	    for (call = AtmCalls; call; call =  call->next)
		if (call->sd > 0)
		    FD_SET(call->sd, &set);

	    error = select(fds, &set, NULL, NULL, next_timer());
	    if (error < 0) {
		if (errno != EINTR) {
		    perror("select");
		    /* abort(); */
		    
		    if (kill(pid, SIGTERM) < 0) /* kill the child  */
			perror("kill");
		    abort();
		}
		continue;
	    }
	    
	    gettimeofday(&now, NULL);

	    /* have a look at PIPE messages */
	    if (FD_ISSET(psd[0], &set))
		fin = process_cmd();

	    /* have a look at signaling messages */
	    if (FD_ISSET(sig_sd, &set)) {
		if ((error = recv_atmsigd(&type)) < 0) {
		    if (error == -2) { 
			/* sigd has closed its tcp socket with client */
			sig_down = 1;
			diag(COMPONENT, DIAG_ERROR, 
			     "\n\nmain_loop: !!! CONNECTION WITH ATMSIGD "
			       "IS DOWN !!!");
			diag(COMPONENT, DIAG_ERROR, 
			     "TEARING DOWN CURRENT ATM CALLS");
			(void) close_all(sig_down);
			(void) close(sig_sd);
		    }
		    if ((error == -2) || (error == -10))
			if (kill(pid, SIGTERM) < 0) /* kill the child  */
			    perror("kill");
		    return(-2);
		}
	    }

	    /* have a look at what's coming on p2p and CCVC */
	    for (call = AtmCalls; call; call =  call->next) {
		sd = call->sd;
		if ((sd > 0) && (FD_ISSET(sd, &set))) {
		    if ((sd == p2p_sd) || (sd == cvc_sd)) {

			error = recv_mars(sd);
			if (error != 0)
			    diag(COMPONENT, DIAG_ERROR, 
				 "main_loop: failed to read message from "
				 "mars");
		    } else {
			static long abuffer[MAX_BUF_LEN / sizeof(long) + 1];
			u_int8_t *buffer = (u_int8_t *)abuffer;

			diag(COMPONENT, DIAG_WARN, 
			     "main_loop: unexpected data");
			/* discard the packet */
			(void) read(sd, buffer, MAX_BUF_LEN);
		    }
		}
	    }

	    /* have a look at routing socket */
	    if (FD_ISSET(rt_sd, &set)) {
		diag(COMPONENT, DIAG_DEBUG, 
		     "main_loop: incoming routing message");
		error = process_routing_msg(rt_sd);
		if (error != 0)
		    diag(COMPONENT, DIAG_WARN, 
			 "main_loop: got trouble with routing message");
	    }   


	    /* expire timers after handling messges to make sure we don't
	       time out unnecessarily because of scheduling delays */
	    expire_timers();
	    
	}
    }
    return(0);
}

/* ------------------------------- main ...  ------------------------------- */

static void usage(const char *name)
{
    fprintf(stderr,
	    "usage: %s [ -s selector ] \n", name);
    exit(1);
}

int main(int argc, char **argv)
{
    int c, error;
    char *end;
    long long_sel;
    u_int8_t selector;

    set_application("CLMB");
    dump_dir = NULL;
    set_verbosity(NULL, DIAG_DEBUG);

/* initialize default selector value */
    selector = 0x00;

    if (argc > 1) {
	while ((c = getopt(argc, argv, "s:")) != EOF)
	    switch (c) {
	    case 's':		/* choose a selector for local ATM address */
		long_sel = strtol(optarg, &end, 0);
		if (*end)
		    usage(argv[0]);
		if (long_sel > 0xff) {
		    diag(COMPONENT, DIAG_ERROR,
			 "Bad selector value!  0x00 <= selector <= 0xff");
		    exit(-1);
		}
		selector = (u_int8_t) long_sel;
		break;
	    default:
		usage(argv[0]);
	    }
    }

    diag(COMPONENT, DIAG_INFO, "\nATM signaling client "
#ifdef UNI30
	 "UNI 3.0"
#endif
#ifdef UNI31
	 "UNI 3.1"
#ifdef ALLOW_UNI30
	 "+3.0compat"
#endif
#endif
#ifdef UNI40
	 "UNI 4.0"
#ifdef Q2963_1
	 "+Q.2931.1"
#endif
#endif
	 "/AAL5, version " VERSION "\n");
  
    init_current_time();
    init_addr(myitf, selector);
    CurrVcc = 0;
    
    if ((error = register_with_atmsigd()) != 0) {
	diag(COMPONENT, DIAG_ERROR, "main: failed to register with atmsigd");
	exit(error);
    }
    diag(COMPONENT, DIAG_DEBUG, "\tWaiting for incoming messages.....");

    /* initialize cluster member */
    init_mars_client(clmb_type);
    error = main_loop();
    diag(COMPONENT, DIAG_DEBUG, "deregistering with atmsigd");
    deregister_with_atmsigd();
    (void) close(sig_sd);
    diag(COMPONENT, DIAG_DEBUG, "End of session");
    return (error);
}
