/* clmb_menu.c - interface menu for MARS Cluster Member */

/* Written 1999 by Mohsen.Souissi@inria.fr, INRIA */

#include <sys/types.h>
#include <sys/socket.h>
#include <net/if.h>
#include <netinet/in.h>
#include <net/route.h>
#include <netnatm/natm.h>
#include <arpa/inet.h>
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>
#include <signal.h>
#include <limits.h>
#include <sys/wait.h>
#include <sys/queue.h>

#include "../linux-atm.h"
#include "atmd.h"
#include "atm.h"
#include "ipv6_tools.h"
#include "misc_tools.h"

#include "atmcall.h"
#include "scl_io.h"

#include "mars_msg.h"
#include "tables.h"

#include "mcl_tab.h"
#include "mcl_io.h"
#include "to_mars.h"


#define COMPONENT	"CLMB_MENU"

int psd[2];			/* socket pair descriptors */

enum menu_choice {
    ch_show_calls = 1, 
    ch_memb = 2, ch_reg = 21, ch_show_memb = 22, 
    ch_join = 23, ch_leave = 24, ch_dereg = 25,
    ch_res = 3, ch_show_res = 31, ch_res_maddr = 32, ch_rel_maddr = 33, 
    ch_p2m = 4, ch_show_p2m = 41, ch_setup = 42, ch_refresh = 43,
    ch_close = 44, ch_close_all_p2m = 45,
    ch_close_all = 5, ch_quit = 6
}; 
    

static void menu_display(void) 
{
    (void) sleep(1);
    printf("\n\n");
    printf("\t\t****************************************\n");
    printf("\t\t*       Custer Member Main MENU        *\n");
    printf("\t\t****************************************\n\n\n");
    printf("1)  Show ATM Calls\n");
    printf("2)  Membership Management\n");
    printf("3)  Multicast Address Resolution\n");
    printf("4)  Point-to-Multipoint ATM Call Management\n");
    printf("5)  Close All ATM Calls\n");
    printf("6)  Quit\n");
    printf("\nchoice: ");
}

static int input_membership(void)		/* Membership Management */
{
    int ret;
    int num, choice, len;
    char choice_str[BUF_LEN];
    u_int8_t type;


  sub_menu:
    printf("\n\n\n\n");
    printf("\t---------------------------------------------\n");
    printf("\t*           Membership Management           *\n");
    printf("\t---------------------------------------------\n\n");
    printf("1) Register Cluster Member\n");
    printf("2) Show Membership Info\n");
    printf("3) Join Multicast Group\n");
    printf("4) Leave Multicast Group\n");
    printf("5) Deregister Cluster Member\n");
    printf("6) Back to MENU\n");
    
    printf("\nchoice: ");
    if (fgets(choice_str, sizeof(choice_str), stdin) == NULL) {
	diag(COMPONENT, DIAG_ERROR, "I/O ERROR");
	abort();		/* something is going wrong */
    }

    if ((len = strlen(choice_str)) < 2)
	goto sub_menu;
    choice_str[len - 1] = '\0';	/* replacing '\n' */
    type = ch_memb;
    if ((num = str2num(choice_str)) < 0) {
	diag(COMPONENT, DIAG_ERROR, 
	     "input_membership: str2num: %s : not a number", choice_str);
	goto sub_menu; 
    }
    else {
	choice = num;
	switch (choice) {
	case 1:
	case 2:
	case 3:
	case 4:
	case 5:
	    type = 10*type + num;
	    break;

	case 6:
	    return(1);		/* will go back to menu_display */

	default:
	    diag(COMPONENT, DIAG_ERROR, 
		 "input_membership: %d BAD CHOICE", choice);
	    goto sub_menu;
	}
    }
    
    if ((type == ch_join) || (type == ch_leave)) {
	struct child_pmsg wmsg;

	/* initialize pipe message and fill in it */
	bzero(&wmsg, sizeof(wmsg));
	wmsg.type = type;

	ret = get_mgroup_addr(&wmsg.ip6m);
	if (ret != 0)
	    goto sub_menu;
	if ((ret = write(psd[1], &wmsg, sizeof(wmsg))) < 0) {
	    diag(COMPONENT, DIAG_ERROR, 
		 "input_membership write: %s", strerror(errno));
	    return(ret);
	}
	if (ret != sizeof(wmsg)) {
	    diag(COMPONENT, DIAG_ERROR, "input_membership : bad write");
	    return(1);
	}
    } else {
	if ((ret = write(psd[1], &type, 1)) < 0) {
	    diag(COMPONENT, DIAG_ERROR, 
		 "input_membership write: %s", strerror(errno));
	    return(ret);
	}
	if (ret != 1) {
	    diag(COMPONENT, DIAG_ERROR, 
		 "input_membership (case %d): bad write", type);
	    return(ret);
	}
    }

    return(0);
}

static int input_resolution(void)	/* Multicast Address Resolution */
{
    int ret;
    int num, choice, len;
    char choice_str[BUF_LEN];
    u_int8_t type;


  sub_menu:
    printf("\n\n\n\n");
    printf("\t---------------------------------------------\n");
    printf("\t*       Multicast Address Resolution        *\n");
    printf("\t---------------------------------------------\n\n");
    printf("1) Show Resolved Multicast Groups\n");
    printf("2) Resolve Multicast Group\n");
    printf("3) Release Resolved Multicast Group\n");
    printf("4) Back to MENU\n");
    
    printf("\nchoice: ");
    if (fgets(choice_str, sizeof(choice_str), stdin) == NULL) {
	diag(COMPONENT, DIAG_ERROR, "I/O ERROR");
	abort();		/* something is going wrong */
    }

    if ((len = strlen(choice_str)) < 2)
	goto sub_menu;
    choice_str[len - 1] = '\0';	/* replacing '\n' */
    type = ch_res;
    if ((num = str2num(choice_str)) < 0) {
	diag(COMPONENT, DIAG_ERROR, 
	     "input_resolution: str2num: %s : not a number", choice_str);
	goto sub_menu; 
    }
    else {
	choice = num;
	switch (choice) {
	case 1:
	case 2:
	case 3:
	    type = 10*type + num;
	    break;

	case 4:
	    return(1);		/* will go back to menu_display */

	default:
	    diag(COMPONENT, DIAG_ERROR, 
		 "input_resolution: %d BAD CHOICE", choice);
	    goto sub_menu;
	}
    }
    
    if ((type == ch_res_maddr) || (type == ch_rel_maddr)) {
	struct child_pmsg wmsg;

	/* initialize pipe message and fill in it */
	bzero(&wmsg, sizeof(wmsg));
	wmsg.type = type;

	ret = get_mgroup_addr(&wmsg.ip6m);
	if (ret != 0)
	    goto sub_menu;
	if ((ret = write(psd[1], &wmsg, sizeof(wmsg))) < 0) {
	    diag(COMPONENT, DIAG_ERROR, 
		 "input_resolution write: %s", strerror(errno));
	    return(ret);
	}
	if (ret != sizeof(wmsg)) {
	    diag(COMPONENT, DIAG_ERROR, "input_resolution : bad write");
	    return(1);
	}
    } else {
	if ((ret = write(psd[1], &type, 1)) < 0) {
	    diag(COMPONENT, DIAG_ERROR, 
		 "input_resolution write: %s", strerror(errno));
	    return(ret);
	}
	if (ret != 1) {
	    diag(COMPONENT, DIAG_ERROR, 
		 "input_resolution (case %d): bad write", type);
	    return(ret);
	}
    }

    return(0);
}

/* Setup a new p2m ATM call */
static int input_setup_p2m(struct child_pmsg *wmsg)
{
    int fw_traffic_class, max_fw_pcr, max_fw_sdu;
    int ret, len;
    char choice_str[BUF_LEN];
 
    ret = get_mgroup_addr(&wmsg->ip6m);
    if (ret != 0)
	return(-1);
 
    printf("\nEnter QoS and Traffic Parameters:\n");
 
    printf("\nForward Traffic Class: [1=UBR / 2=CBR] [default=UBR]: ");
 
    if (fgets(choice_str, sizeof(choice_str), stdin) == NULL) {
	diag(COMPONENT, DIAG_ERROR, "I/O ERROR");
	abort();		/* something is going wrong */
    }

    if ((len = strlen(choice_str)) < 2) {
        fw_traffic_class = ATM_UBR;
        goto retry_fw_pcr;
    }
    choice_str[len - 1] = '\0'; /* replacing '\n' */
    if ((fw_traffic_class = str2num(choice_str)) < 0) {
        printf("\n input_setup_p2m: str2num: %s : not a number", choice_str);
        return(-1);
    }
    if (fw_traffic_class > 2) {
        printf("\n input_setup_p2m: illegal Traffic Class value %s", 
	       choice_str);
        return(-1);
    }
 
 retry_fw_pcr:
    printf("\nMaximum Forward PCR: [default = ATM_OC3_PCR/10 = 35320]: ");
    if (fgets(choice_str, sizeof(choice_str), stdin) == NULL) {
	diag(COMPONENT, DIAG_ERROR, "I/O ERROR");
	abort();		/* something is going wrong */
    }

    if ((len = strlen(choice_str)) < 2) {
        max_fw_pcr = ATM_OC3_PCR/10;		/* cf RFC 1755, par. 7.1 */
        goto retry_fw_sdu;
    }
    choice_str[len - 1] = '\0'; /* replacing '\n' */
    if ((max_fw_pcr = str2num(choice_str)) < 0) {
        printf("\n input_setup_p2m: str2num: %s : not a number", choice_str);
        return(-1);
    }
    if (max_fw_pcr > ATM_OC3_PCR) {
        printf("\n input_setup_p2m: illegal PCR value %s", choice_str);
        return(-1);
    }
 
 retry_fw_sdu:
    printf("\nForward SDU size (in bytes): [default = 9188]: ");
    if (fgets(choice_str, sizeof(choice_str), stdin) == NULL) {
	diag(COMPONENT, DIAG_ERROR, "I/O ERROR");
	abort();		/* something is going wrong */
    }

    if ((len = strlen(choice_str)) < 2) {
        max_fw_sdu = DEFAULT_MAX_SDU;
        goto fill_in;
    }
    choice_str[len - 1] = '\0'; /* replacing '\n' */
    if ((max_fw_sdu = str2num(choice_str)) < 0) {
        printf("\n input_setup_p2m: str2num: %s : not a number", choice_str);
        return(-1);
    }
    if (max_fw_sdu > MAX_SDU_SIZE) {
        printf("\n input_setup_p2m: too big SDU size value: %s > %d",
               choice_str, MAX_SDU_SIZE);
        return(-1);
    }
 
 fill_in:
    /* carry on filling in child_pmsg structure */
    wmsg->fw_traffic_class = (u_int8_t) fw_traffic_class;
    wmsg->max_fw_pcr = max_fw_pcr;
    wmsg->max_fw_sdu = max_fw_sdu;

    return(0);
}

static int input_p2m(void)	/* p2m call management */
{
    int ret;
    int num, choice, len;
    char choice_str[BUF_LEN], vcc_str[BUF_LEN];
    struct child_pmsg wmsg;
    u_long vcc;
    u_int8_t type;


  sub_menu:
    printf("\n\n\n\n");
    printf("\t---------------------------------------------\n");
    printf("\t*  Point-to-Multipoint ATM Call Management  *\n");
    printf("\t---------------------------------------------\n\n");
    printf("1) Show p2m ATM Calls\n");
    printf("2) Setup a New p2m Call\n");
    printf("3) Refresh a p2m Call\n");
    printf("4) Close a p2m Call\n");
    printf("5) Close All p2m Calls\n");
    printf("6) Back to MENU\n");

    printf("\nchoice: ");
    if (fgets(choice_str, sizeof(choice_str), stdin) == NULL)  {
	diag(COMPONENT, DIAG_ERROR, "I/O ERROR");
	abort();		/* something is going wrong */
    }

    if ((len = strlen(choice_str)) < 2)
	goto sub_menu;
    choice_str[len - 1] = '\0';	/* replacing '\n' */
    type = ch_p2m;
    if ((num = str2num(choice_str)) < 0) {
	diag(COMPONENT, DIAG_ERROR, 
	     "input_p2m: str2num: %s : not a number", choice_str);
	goto sub_menu; 
    }
    else {
	choice = num;
	switch (choice) {
	case 1:
	case 2:
	case 3:
	case 4:
	case 5:
	    type = 10*type + num;
	    break;

	case 6:
	    return(1);		/* will go back to menu_display */

	default:
	    diag(COMPONENT, DIAG_ERROR, "input_p2m: %d BAD CHOICE", choice);
	    goto sub_menu;
	}
    }

    if ((type == ch_setup) || (type == ch_refresh) || (type == ch_close)) {    
	/* initialize pipe message and fill in it */
	bzero(&wmsg, sizeof(wmsg));
	wmsg.type = type;
    
	switch (type) {
	case ch_setup:
	    if (input_setup_p2m(&wmsg) != 0) {
		diag(COMPONENT, DIAG_ERROR, "input_p2m: TRY AGAIN !");
		goto sub_menu;
	    }
	    break;
	case ch_refresh:
	case ch_close: 
	retry_vcc:
	    printf("\n vcc : ");
	    if (fgets(vcc_str, sizeof(vcc_str), stdin) == NULL) {
		diag(COMPONENT, DIAG_ERROR, "I/O ERROR");
		abort();		/* something is going wrong */
	    }

	    if ((len = strlen(vcc_str)) < 2) 
		goto retry_vcc;
	    vcc_str[len - 1] = '\0'; /* replacing '\n' */
	    if ((ret = str2num(vcc_str)) < 0) {
		diag(COMPONENT, DIAG_ERROR, 
		     "input_p2m: %s : Not a number", vcc_str);
		fflush(stdout);
		(void) sleep(1);
		goto sub_menu;
	    }
	    vcc = (u_long) ret;
	    wmsg.vcc = vcc;
	    break;
    default:		/* should never happen... already filtered */
	diag(COMPONENT, DIAG_ERROR, "input_p2m: %d unknown type", type);
	return(-1);
	}

	if ((ret = write(psd[1], &wmsg, sizeof(wmsg))) < 0) {
	    diag(COMPONENT, DIAG_ERROR, 
		 "input_p2m write: %s", strerror(errno));
	    return(ret);
	}
	if (ret != sizeof(wmsg)) {
	    diag(COMPONENT, DIAG_ERROR, "input_p2m : bad write");
	    return(1);
	}
    } else {
	if ((ret = write(psd[1], &type, 1)) < 0) {
	    diag(COMPONENT, DIAG_ERROR, 
		 "input_p2m write: %s", strerror(errno));
	    return(ret);
	}
	if (ret != 1) {
	    diag(COMPONENT, DIAG_ERROR, 
		 "input_p2m (case %d): bad write", type);
	    return(ret);
	}
    }

    return(0);
}

int user_input(void)
{
    int ret = 0, len, choice;
    char choice_str[BUF_LEN];
    u_int8_t type, ret_code;

 menu:
    menu_display();
    if (fgets(choice_str, sizeof(choice_str), stdin) == NULL) {
	diag(COMPONENT, DIAG_ERROR, "I/O ERROR");
	abort();		/* something is going wrong */
    }
    if ((len = strlen(choice_str)) < 2) 
	goto menu;
    choice_str[len - 1] = '\0';	/* replacing '\n' */
    if ((ret = str2num(choice_str)) < 0) {
	diag(COMPONENT, DIAG_ERROR, 
	     "user_input: str2num: %s : Not a number", choice_str);
	fflush(stdout);
	(void) sleep(2);
	goto menu;
    }
    choice = ret;
    type = (u_int8_t) choice;
    switch (choice) {
    case ch_show_calls:	/* (1) Show ATM Calls */
    case ch_close_all:	/* (5) Close All ATM Calls */
	if ((ret = write(psd[1], &type, 1)) < 0) { /* send only the type */
	    diag(COMPONENT, DIAG_ERROR, 
		  "user_input write: %s", strerror(errno));
	    goto menu;
	}
	if (ret != 1) {
	    diag(COMPONENT, DIAG_ERROR, 
		 "user_input (case %d): bad write", type);
	    goto menu;
	}
	if ((ret = read(psd[1], &ret_code, 1)) != 1) {
	    diag(COMPONENT, DIAG_ERROR, 
		 "user_input: read error %s", strerror(errno));
	    goto menu;
	}
	if (ret_code != 0)
	    diag(COMPONENT, DIAG_WARN, 
		 "user_input: got trouble when processing command (%d)", 
		 type);
	goto menu;
	
    case ch_memb:	/* (2) Membership Management */
	if (input_membership() != 0)
	    goto menu;
	if ((ret = read(psd[1], &ret_code, 1)) != 1) {
	    diag(COMPONENT, DIAG_ERROR, 
		 "user_input: read error %s", strerror(errno));
	    goto menu;
	}
	if (ret_code != 0)
	    diag(COMPONENT, DIAG_WARN, 
		 "user_input: got trouble when processing command (%d)", 
		 type);
	goto menu;
	
    case ch_res:	/* (3) Multicast Address Resoultion mgroup */
	if (input_resolution() != 0)
	    goto menu;
	if ((ret = read(psd[1], &ret_code, 1)) != 1) {
	    diag(COMPONENT, DIAG_ERROR, 
		 "user_input: read error %s", strerror(errno));
	    goto menu;
	}
	if (ret_code != 0)
	    diag(COMPONENT, DIAG_WARN, 
		 "user_input: got trouble when processing command (%d)", 
		 type);
	goto menu;
	
    case ch_p2m:	/* (4) Point-to-Multipoint Calls */
	if (input_p2m() != 0)
	    goto menu;
	if ((ret = read(psd[1], &ret_code, 1)) != 1) {
	    diag(COMPONENT, DIAG_ERROR, 
		 "user_input: read error %s", strerror(errno));
	    goto menu;
	}
	if (ret_code != 0)
	    diag(COMPONENT, DIAG_WARN, 
		 "user_input: got trouble when processing command (%d)", 
		 type);
	goto menu;
	
    case ch_quit:	/* (6) QUIT */
	if ((ret = write(psd[1], &type, 1)) < 0) {
	    diag(COMPONENT, DIAG_ERROR, 
		  "user_input write: %s", strerror(errno));
	    goto menu;
	}
	if (ret != 1) {
	    diag(COMPONENT, DIAG_ERROR, 
		 "user_input (case %d): bad write", type);
	    goto menu;
	}
	if ((ret = read(psd[1], &ret_code, 1)) != 1) {
	    diag(COMPONENT, DIAG_ERROR, 
		 "user_input: read error %s", strerror(errno));
	    goto menu;
	}
	switch(ret_code) {
	case 0:
	    return(1);		/* we are about to quit */
	case 1:
	    goto menu;
	default:
	    diag(COMPONENT, DIAG_WARN, 
		 "user_input: got trouble when processing command (%d)", 
		   type);
	    goto menu;
	}
    default:
	diag(COMPONENT, DIAG_ERROR, "user_input: %d BAD CHOICE", choice);
	fflush(stdout);
	(void) sleep(2);
	goto menu;
    }	
}

static int process_p2m_setup(struct child_pmsg *rmsg)
{
    struct in6_addr ip6m;
    struct resolv_entry *res_e = NULL, *new_res_e;
    struct atm_qos *qos;
    char v[64];
    int error;

    bcopy(&rmsg->ip6m, &ip6m, IN6_ADDR_LEN);
    diag(COMPONENT, DIAG_DEBUG,
	"Attempting to set up a p2m VC to IPv6 Multicast Address: %s", 
	inet_ntop(AF_INET6, &ip6m, v, sizeof(v)));

    /* allocate and fill in qos structure */
    qos = alloc_t(struct atm_qos);
    bzero(qos, sizeof(*qos));
    qos->txtp.traffic_class = rmsg->fw_traffic_class;
    qos->txtp.max_pcr = rmsg->max_fw_pcr;
    qos->txtp.max_sdu = rmsg->max_fw_sdu;
    qos->aal = ATM_AAL5;

    error = resolve_mgroup(&res_e, &ip6m, RTM_ADD);
    if (error != 0) {
	diag(COMPONENT, DIAG_ERROR,
	     "process_p2m_setup: got trouble when resolving mgroup. Retry it!");
	return(-1);
    }

    /* if no resolve entry is available or the entry isn't in  "valid" state */
    if ((res_e == NULL) || (res_e->rstate != rs_valid)) {
	diag(COMPONENT, DIAG_DEBUG, "Waiting for MARS_MULTI Message");

	if (qos->txtp.traffic_class != ATM_UBR) {
	    printf("\n!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!\n");
	    printf("process_p2m_setup: NOT READY YET. ");
	    printf("JUST COME BACK LATER AND RETRY IT!!!\n");
	    printf("!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!\n");
	}
	return(0);
    }
    /* now, a corresponding resolve entry is available */

    /* don't allow more than one Best Effort p2m VC */
    
    if (qos->txtp.traffic_class == ATM_UBR) {
	diag(COMPONENT, DIAG_WARN, 
	     "process_p2m_setup: default Multicast VC already exists !!!");
	return(0);
    }

    /* 
     * If the default p2m VC contains only 1 leaf, the group may be 
     * MCS-supported. In such a case QoS may be irrelevent, send a warning
     */
    if (res_e->mvc->nbactp == 1) {
	 printf("\n!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!\n");
	 diag(COMPONENT, DIAG_WARN, 
	     "This Group may be supported by a MCS, end-to-end QoS "
	      "might not work !!! Check for MCS existence for this group !");
	 printf("\n!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!\n");
     }

    /* duplicate the resolve entry and build corresponding p2m VC (with QoS) */

    new_res_e = dup_resolv_entry(res_e, qos);
    error = build_mvc(new_res_e);
    if (error != 0) {
	diag(COMPONENT, DIAG_ERROR, 
	     "process_p2m_setup: failed to setup p2m ATM Call");
	return(-1);
    }

    return(0);
}

static int refresh_p2m_call(ATMCALL *call)
{
    struct resolv_entry *res_e, *my_res_e;
    struct ept_head *to_drop_head, *to_add_head, *head;
    struct ept_entry *walk, *new_ept_e, *next, *copy, *last = NULL;
    PARTY *party;
    int error, i;

    error = resolve_mgroup(&res_e, &call->rin6, RTM_ADD);
    if (error != 0) {
	diag(COMPONENT, DIAG_ERROR,
	     "refresh_p2m_call: got trouble when resolving mgroup. Retry it!");
	return(-1);
    }

    /* if no resolve_entry is available or the entry isn't in "valid" state */
    if ((res_e == NULL) || (res_e->rstate != rs_valid)) {
	diag(COMPONENT, DIAG_DEBUG, "Waiting for MARS_MULTI Message");
	printf("\n!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!\n");
	printf("refresh_p2m_call: NOT READY YET. ");
	printf("JUST COME BACK LATER AND RETRY IT!!!\n");
	printf("!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!\n");
	return(0);
    }
    /* now, a corresponding resolve entry is available */

    /* do as if we were revalidating a default p2m call */
    my_res_e = call->res_e;
    for (walk = LIST_FIRST(&res_e->ept_head);
	 walk;
	 walk = LIST_NEXT(walk, links)) {
	new_ept_e = new_ept_entry(&walk->addr, my_res_e);
	LIST_INSERT_HEAD(&my_res_e->new_head, new_ept_e, links);
    }
    /* build "drop" and "add" lists */
    to_drop_head = alloc_t(struct ept_head);
    LIST_INIT(to_drop_head);
    
    build_drop_add_lists(my_res_e, to_drop_head);
    to_add_head = &my_res_e->new_head;
    head = &my_res_e->ept_head;

    /* drop obsolete parties */
    for (walk = LIST_FIRST(to_drop_head); walk; walk = next) {
	next = LIST_NEXT(walk, links);
	if (next == NULL) {
	    /* 
	     * it's the last party, don't remove it now but mark it 
	     * and drop it later
	     */
	    last = walk;
	    break;
	}
	party = walk->party;
	if (party != NULL) {
	    error = active_drop_party(0, 0, party);
	    if (error != 0) {
		diag(COMPONENT, DIAG_ERROR, 
		     "refresh_p2m_call: failed to initiate dropping leaf: ");
		for (i = 0; i < ATM_ESA_LEN; i++)
		    printf("%02X", walk->addr.prv[i]);
		printf("\n");
		/* rollback */
		LIST_INSERT_HEAD(head, walk, links);
		/* mark something has gone wrong */
		error = -1;
	    }
	} else {
	    diag(COMPONENT, DIAG_WARN, 
		 "refresh_p2m_call: no party matching ept_e: ");
	    for (i = 0; i < ATM_ESA_LEN; i++)
		printf("%02X", walk->addr.prv[i]);
	    printf("\n");
	}
    }
    
    /* add new parties */
    for (walk = LIST_FIRST(to_add_head); walk; walk = LIST_NEXT(walk, links)) {
	party = active_add_party(my_res_e->mvc, &walk->addr, NULL, walk, NULL);
	if (party == NULL) {
	    diag(COMPONENT, DIAG_ERROR, 
		 "refresh_p2m_call: failed to initiate adding leaf : ");
	    for (i = 0; i < ATM_ESA_LEN; i++)
		printf("%02X", walk->addr.prv[i]);
	    printf("\n");
	    /* mark something has gone wrong */
	    error = -1;
	} else {
	    /* copy this new ept entry to the resolve list */
	    copy = alloc_t(struct ept_entry);
	    bcopy(walk, copy, sizeof(struct ept_entry));
	    party->ept_e = copy;
	    LIST_INSERT_HEAD(head, copy, links);
	}
    }
    
    /* go back to the remaining party to drop (if any) */
    if (last != NULL) {
	party = last->party;
	if (party != NULL) {
	    error = active_drop_party(0, 0, party);
	    if (error != 0) {
		diag(COMPONENT, DIAG_ERROR, 
		     "refresh_p2m_call: failed to initiate "
		     "dropping leaf : ");
		for (i = 0; i < ATM_ESA_LEN; i++)
		    printf("%02X", last->addr.prv[i]);
		printf("\n");
		/* rollback */
		LIST_INSERT_HEAD(head, last, links);
		/* mark something has gone wrong */
		error = -1;
	    }
	} else {
	    diag(COMPONENT, DIAG_WARN, "no such party: ");
	    for (i = 0; i < ATM_ESA_LEN; i++)
		printf("%02X", last->addr.prv[i]);
	    printf("\n");
	}
    }
    
    /* free useless lists */
    free_ept_entries(to_drop_head);
    free(to_drop_head);
    free_ept_entries(to_add_head);

    return(error);
}


int process_cmd(void)
{
    int ret = 0, len;
    u_int8_t type, ret_code;
    struct child_pmsg rmsg;
    ATMCALL *call;
    pid_t pid;
    int status;
    char str_addr[64];


    if ((ret = read(psd[0], &type, 1)) != 1) {
	diag(COMPONENT, DIAG_ERROR, 
	     "process_cmd: read error %s", strerror(errno));
	ret_code = 0;
	goto feedback;
    }
  
    switch (type) {
    case ch_join:
    case ch_leave:
    case ch_res_maddr:
    case ch_rel_maddr:
    case ch_setup:
    case ch_refresh:
    case ch_close:
	len = sizeof(rmsg);
	bzero(&rmsg, len);
	rmsg.type = type;
	if ((ret = read(psd[0], (u_int8_t *)&rmsg + 1, len - 1)) != len - 1) {
	    diag(COMPONENT, DIAG_ERROR, 
		 "process_cmd: read error %s", strerror(errno));
	    ret_code = 0;
	    goto feedback;
	}
    default:
	break;
    }

    printf("\n\n");
    switch (type) {
    case ch_show_calls:	/* (1) Show Current ATM Calls */
	show_atmcalls();
        ret_code = 0;
        break;

    case ch_reg:	/* (2.1) Register Cluster Member */
	ret = register_mars_client(mars_join);
	ret_code = ret ? 1 : 0;
	break;

    case ch_show_memb:	/* (2.2) Show Membership Info */
	show_membership(clmb_type);
	ret_code = 0;
	break;

    case ch_join: 	/* (2.3) Join Multicast Group */
	/* is interface already member of this group ? */
	if (belongs_to_block_list(&rmsg.ip6m, &Minmax_Table)) {
	    diag(COMPONENT, DIAG_DEBUG, 
		 "process_cmd: interface is already member of mgroup %s "
		 "No message sent",
		 inet_ntop(AF_INET6, &rmsg.ip6m, str_addr, 64));
	    ret_code = 0;
	    break;
	}
	ret = join_serve_mgroup(&rmsg.ip6m, mars_join, FL_LAYER3GRP);
  	ret_code = ret ? 1 : 0;
	break;

    case ch_leave: 	/* (2.4) Leave Multicast Group */
	/* is interface already member of this group ? */
	if (!belongs_to_block_list(&rmsg.ip6m, &Minmax_Table)) {
	    diag(COMPONENT, DIAG_DEBUG, 
		 "process_cmd: interface is not a member of mgroup %s "
		 "No message sent",
		 inet_ntop(AF_INET6, &rmsg.ip6m, str_addr, 64));
	    ret_code = 0;
	    break;
	}

	ret = leave_unserve_minmax(&rmsg.ip6m, &rmsg.ip6m, 
				   mars_leave, FL_LAYER3GRP);
  	ret_code = ret ? 1 : 0;
	break;

    case ch_dereg:	/* (2.5) de-Register Cluster Member */
	ret = signoff_cluster();
	ret_code = ret ? 1 : 0;
	break;
	
    case ch_show_res:	/* (3.1) Show Resolved Multicast Groups */
	show_resolved_mgroups();
	ret_code = 0;
	break;

    case ch_res_maddr:	/* (3.2) Resolve Multicast Group */
	ret = resolve_mgroup(&Curr_Resolved, &rmsg.ip6m, RTM_ADD);
	ret_code = ret ? 1 : 0;
	break;

    case ch_rel_maddr:	/* (3.3) Release Resolved Multicast Group */
	ret = release_resolved_mgroup(&rmsg.ip6m);
	ret_code = ret ? 1 : 0;
	break;

    case ch_show_p2m:	/* (4.1) Show p2m ATM Calls */
	show_p2m_calls();
	ret_code = 0;
	break;

    case ch_setup:	/* (4.2) Setup a New p2m Call */
	ret = process_p2m_setup(&rmsg);
	ret_code = ret ? 1 : 0;
	break;

    case ch_refresh:	/* (4.3) Refresh a p2m Call */
	/* check it is a p2m call */
	if ((call = atmcall_lookup(rmsg.vcc)) == NULL) {
	    diag(COMPONENT, DIAG_ERROR, 
		 "process_cmd: no matching atm call for VCC %ld", rmsg.vcc);
	    ret_code = 1;
	    break;
	}
	if (call->conf != root) {
	    diag(COMPONENT, DIAG_ERROR, 
		 "process_cmd: not a p2m call [VCC = %ld]", call->vcc);
	    ret_code = 1;
	    break;
	}
	if (call->res_e->qos == NULL) {
	    diag(COMPONENT, DIAG_WARN, 
		 "process_cmd: this is the default (Best Effort) p2m VC !!! "
		 "No further action.");
	    ret_code = 0;
	    break;
	}
	ret = refresh_p2m_call(call);
	ret_code = ret ? 1 : 0;
	break;
	

    case ch_close:	/* (4.4) Close a p2m Call */
	/* check it is a p2m call */
	if ((call = atmcall_lookup(rmsg.vcc)) == NULL) {
	    diag(COMPONENT, DIAG_ERROR, 
		 "process_cmd: no matching atm call for VCC %ld", rmsg.vcc);
	    ret_code = 1;
	    break;
	}
	if (call->conf != root) {
	    diag(COMPONENT, DIAG_ERROR, 
		 "process_cmd: not a p2m call [VCC = %ld]", call->vcc);
	    ret_code = 1;
	    break;
	}
	ret = active_close_atmcall(0, call->vcc, call);
	ret_code = ret ? 1 : 0;
	break;

    case ch_close_all_p2m:	/* (4.5) Close All p2m Calls */
	ret = close_all_p2m_calls();
	ret_code = ret ? 1 : 0;
	break;

    case ch_close_all:	/* (5) Close All ATM Calls */
	ret = close_all(0);
	ret_code = ret ? 1 : 0;
	break;
    case ch_quit:	/* (6) Quit */ 
	if (AtmCalls != NULL) {
	    diag(COMPONENT, DIAG_WARN, 
		 "process_cmd: !!! STILL %d PENDING CALL(S)", NbCalls);
	    diag(COMPONENT, DIAG_WARN, "==> GO BACK TO MENU AND DE-REGISTER");
	    (void) sleep(2);
	    ret_code = 1;
	    break;
	}
	ret_code = 0;
	break;
    default:
	diag(COMPONENT, DIAG_ERROR, 
	     "process_cmd: Unknown type of command: %d", type);
	ret_code = 1;
	fflush(stdout);
	(void) sleep(2);
    }

 feedback:
    if ((ret = write(psd[0], &ret_code, 1)) < 0) {
	diag(COMPONENT, DIAG_ERROR, 
	      "process_cmd write: %s", strerror(errno));
	return(-1);
    }
    if ((type == ch_quit) && (ret_code == 0)) {
	diag(COMPONENT, DIAG_DEBUG, "\n!!! Terminating session");
	if ((pid = wait(&status)) < 0) {
	    diag(COMPONENT, DIAG_ERROR, "wait error: %s", strerror(errno));
	    return(-1);
	}
	return(1);
    }
    return(0);
}

