/* marsd_io.c - MARS Daemon I/O operations */

/* Written 1999 by Mohsen.Souissi@inria.fr, INRIA */

#include <sys/param.h>
#include <sys/socket.h>
#include <sys/file.h>
#include <sys/ioctl.h>

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>

#include <net/if.h>
#ifndef if_type
#include <net/if_var.h> 
#endif

#include <net/if_atm.h>
#include <netinet/in.h>
#include <netinet/in6_var.h>
#include <arpa/inet.h>
#include <netnatm/natm.h>

#include <sys/time.h>
#include <sys/types.h>

#include "../linux-atm.h"

#include "atm.h"
#include "atmd.h"
#include "ipv6_tools.h"

#include "atmcall.h"
#include "scl_io.h"

#include "mars_msg.h"
#include "mars_def.h"
#include "tables.h"
#include "marsd_tab.h"
#include "marsd_io.h"

#define COMPONENT "MARSD_IO"



/* 
 * send a mars message to a specific cluster member privately (over its 
 * private p2p VC)
 */
int send_msg_to_mclient(int sd, u_int8_t *msg, u_int msg_len)
{
    int wrote;
   
    if ((wrote = write(sd, msg, msg_len)) == msg_len) {
	NbOutput++;
        return(0);
    }
	
    NbOutputErrors++;
    if (wrote < 0) {
        fprintf(stderr, "to_mclient write: %s", strerror(errno));
        if (errno == EPIPE)
            diag(COMPONENT, DIAG_ERROR, 
		 "send_msg_to_mclient:  MCLIENT [sd = %d] is down", sd);
        return(wrote);
    }
    diag(COMPONENT, DIAG_ERROR, 
	 "bad MCLIENT write: wanted %d, wrote %d", msg_len, wrote);
    return(wrote);
}

/* 
 * send a mars message either on CCVC or on SCVC
 */
int send_msg_to_all(int cvc_sd, u_int8_t *msg, u_int msg_len)
{
    int wrote;
   
    if ((wrote = write(cvc_sd, msg, msg_len)) == msg_len) {
	NbOutput++;
        return(0);
    }

    NbOutputErrors++;
    if (wrote < 0) {
        fprintf(stderr, "send_msg_to_all write: %s", strerror(errno));
        if (errno == EPIPE)
            diag(COMPONENT, DIAG_ERROR, 
		 "send_msg_to_all:  %s is down", 
		 (cvc_sd == ccvc_sd) ? "CCVC" : "SCVC");
        return(wrote);
    }

    diag(COMPONENT, DIAG_ERROR, 
	 "bad  write on %s: wanted %d, wrote %d\n", 
	 (cvc_sd == ccvc_sd) ? "CCVC" : "SCVC", msg_len, wrote);
    return(wrote);
}


int recv_mclient(ATMCALL *mcl_p2p)
{
    static long abuffer[MAX_BUF_LEN / sizeof(long) + 1];
    u_int8_t *buffer = (u_int8_t *)abuffer;
    int size, error, min_len;

    if ((size = read(mcl_p2p->sd, buffer, MAX_BUF_LEN)) < 0) {
	diag(COMPONENT, DIAG_ERROR, "recv mclient: %s", strerror(errno));
	NbInputErrors++;
	return(-1);       
    }
    if (size == 0) {
	diag(COMPONENT, DIAG_ERROR, 
	     "recv mclient: natm_socket [sd = %d] closed by peer", mcl_p2p->sd);
	return(-1);
    }

    NbInput++;    
    min_len =  ATMLLC_LEN + sizeof(struct mars_msg);
    if (size < min_len) {
	diag(COMPONENT, DIAG_ERROR,
	     "mars message is too short (got %d < %d)", size, min_len);
	NbInputErrors++;
	return(-1);
    }
	
    error = dispatch_mars_msg(mcl_p2p, buffer, size);
    return error < 0 ? -1 : 0;
}
