/* marsd_menu.c - interface menu for MARS daemon */

/* Written 1999 by Mohsen.Souissi@inria.fr, INRIA */

#include <sys/types.h>
#include <sys/socket.h>
#include <net/if.h>
#include <netinet/in.h>
#include <netnatm/natm.h>
#include <arpa/inet.h>
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>
#include <signal.h>
#include <limits.h>
#include <sys/wait.h>
#include <sys/queue.h>

#include "../linux-atm.h"
#include "atmd.h"
#include "atm.h"
#include "ipv6_tools.h"
#include "misc_tools.h"

#include "atmcall.h"
#include "scl_io.h"

#include "tables.h"

#include "mars_msg.h"
#include "marsd_tab.h"
#include "marsd_io.h"

#define COMPONENT	"MENU"

int psd[2];			/* socket pair descriptors */

enum menu_choice {
    ch_show_calls = 1, 
    ch_clmb = 2, ch_show_clmb = 21, ch_host_table = 22, ch_dispose_clmb = 23,
    ch_mcs = 3, ch_show_mcs = 31, ch_server_table = 32, ch_dispose_mcs = 33, 
    ch_stats = 4,
    ch_close_all = 5, ch_quit = 6
}; 

static void menu_display(void) 
{
    (void) sleep(1);
    printf("\n\n");
    printf("\t\t****************************************\n");
    printf("\t\t*            MARS Main MENU            *\n");
    printf("\t\t****************************************\n\n\n");
    printf("1) Show ATM Calls\n");
    printf("2) Cluster Member Management\n");
    printf("3) MCS Management\n");
    printf("4) Show MARSD Stats\n");
    printf("5) Close All ATM VCs\n");
    printf("6) Quit\n");
    printf("\nchoice: ");
}

/* MARS Client Management */
static int input_mclient_management(enum mcl_type mtype)
{
    int ret;
    int num, choice, len;
    char choice_str[BUF_LEN], mclid_str[BUF_LEN];
    u_int8_t type;
    u_int16_t mcl_id;
    struct child_pmsg wmsg;

  sub_menu:
    printf("\n\n\n\n");
    printf("\t---------------------------------------------\n");
    printf("\t*           %s Management\t           *\n", mcl_name[mtype]);
    printf("\t---------------------------------------------\n\n");
    printf("1) Show All %ss\n", mcl_name[mtype]);
    printf("2) Show %s Map Table\n", 
	  (mtype == clmb_type) ? "Host" : "Server");
    printf("3) Dispose of %s\n", mcl_name[mtype]);
    printf("4) Back to MENU\n");
    
    printf("\nchoice: ");
    if (fgets(choice_str, sizeof(choice_str), stdin) == NULL) {
	diag(COMPONENT, DIAG_ERROR, "I/O ERROR");
	abort();		/* something is going wrong */
    }

    if ((len = strlen(choice_str)) < 2)
	goto sub_menu;
    choice_str[len - 1] = '\0';	/* replacing '\n' */
    type = (mtype == clmb_type) ? ch_clmb : ch_mcs;
    if ((num = str2num(choice_str)) < 0) {
	diag(COMPONENT, DIAG_ERROR, 
	     "input_mclient_management: str2num: %s : not a number", 
	     choice_str);
	goto sub_menu; 
    }
    else {
	choice = num;
	switch (choice) {
	case 1:
	case 2:
	case 3:
	    type = 10*type + num;
	    break;

	case 4:
	    return(1);		/* will go back to menu_display */

	default:
	    diag(COMPONENT, DIAG_ERROR, 
		 "input_mclient_management: %d BAD CHOICE", choice);
	    goto sub_menu;
	}
    }
    
    if ((type == ch_dispose_clmb) || (type == ch_dispose_mcs)) {
	/* initialize pipe message and fill in it */
	bzero(&wmsg, sizeof(wmsg));
	wmsg.type = type;

	/* Get MARS Client Id (CMI/MCSId) */
    retry_mclid:
	printf("\n%s: ",
	       (mtype == clmb_type) 
	       ? "Cluster Member Identifier (CMI)" : 
	       "MCS Identifier (MCSId)");
	if (fgets(mclid_str, sizeof(mclid_str), stdin) == NULL) {
	    diag(COMPONENT, DIAG_ERROR, "I/O ERROR");
	    abort();		/* something is going wrong */
	}

	if ((len = strlen(mclid_str)) < 2) 
	    goto retry_mclid;
	mclid_str[len - 1] = '\0'; /* replacing '\n' */
	if ((ret = str2num(mclid_str)) < 0) {
	    diag(COMPONENT, DIAG_ERROR, 
		 "input_mclient_management: %s : Not a number", mclid_str);
	    fflush(stdout);
	    (void) sleep(1);
	    goto sub_menu;
	}
	mcl_id = (u_int16_t) ret;
	wmsg.mcl_id = mcl_id;
	
	if ((ret = write(psd[1], &wmsg, sizeof(wmsg))) < 0) {
	    diag(COMPONENT, DIAG_ERROR, 
		 "input_mclient_management write: %s", strerror(errno));
	    return(ret);
	}
	if (ret != sizeof(wmsg)) {
	    diag(COMPONENT, DIAG_ERROR, "input_mclient_management: bad write");
	    return(1);
	}
    } else {
	if ((ret = write(psd[1], &type, 1)) < 0) {
	    diag(COMPONENT, DIAG_ERROR, 
		 "input_mclient_management write: %s", strerror(errno));
	    return(ret);
	}
	if (ret != 1) {
	    diag(COMPONENT, DIAG_ERROR, 
		 "input_mclient_management (case %d): bad write", type);
	    return(ret);
	}
    }

    return(0);
}

int user_input(void)
{
    int ret = 0, len, choice;
    char choice_str[BUF_LEN];
    u_int8_t type, ret_code;
    enum mcl_type mtype;

 menu:
    menu_display();
    if (fgets(choice_str, sizeof(choice_str), stdin) == NULL) {
	diag(COMPONENT, DIAG_ERROR, "I/O ERROR");
	abort();		/* something is going wrong */
    }
    if ((len = strlen(choice_str)) < 2) 
	goto menu;
    choice_str[len - 1] = '\0';	/* replacing '\n' */
    if ((ret = str2num(choice_str)) < 0) {
	diag(COMPONENT, DIAG_ERROR, 
	     "user_input: str2num: %s : Not a number", choice_str);
	fflush(stdout);
	(void) sleep(2);
	goto menu;
    }
    choice = ret;
    type = (u_int8_t) choice;
    switch (choice) {
    case ch_show_calls:		/* (1) Show ATM Calls */
    case ch_stats:		/* (4) Show MARSD Stats */
    case ch_close_all:		/* (5) Close All ATM Calls */
	if ((ret = write(psd[1], &type, 1)) < 0) { /* send only the type */
	    diag(COMPONENT, DIAG_ERROR, 
		 "user_input write: %s", strerror(errno));
	    goto menu;
	}
	if (ret != 1) {
	    diag(COMPONENT, DIAG_ERROR, 
		 "user_input (case %d): bad write", type);
	    goto menu;
	}
	if ((ret = read(psd[1], &ret_code, 1)) != 1) {
	    diag(COMPONENT, DIAG_ERROR, 
		 "user_input: read error %s", strerror(errno));
	    goto menu;
	}
	if (ret_code != 0)
	    diag(COMPONENT, DIAG_WARN, 
		 "user_input: got trouble when processing command (%d)", 
		 type);
	goto menu;

    case ch_clmb:
    case ch_mcs:
	mtype = (type == ch_clmb) ? clmb_type : mcs_type;
	if (input_mclient_management(mtype) != 0)
	    goto menu;
	if ((ret = read(psd[1], &ret_code, 1)) != 1) {
	    diag(COMPONENT, DIAG_ERROR, 
		 "user_input: read error %s", strerror(errno));
	    goto menu;
	}
	if (ret_code != 0)
	    diag(COMPONENT, DIAG_WARN, 
		 "user_input: got trouble when processing command (%d)", 
		 type);
	goto menu;

    case ch_quit:			/* (6) Quit */
	if ((ret = write(psd[1], &type, 1)) < 0) {
	    diag(COMPONENT, DIAG_ERROR, 
		 "user_input write: %s", strerror(errno));
	    goto menu;
	}
	if (ret != 1) {
	    diag(COMPONENT, DIAG_ERROR, 
		 "user_input (case %d): bad write", type);
	    goto menu;
	}
	if ((ret = read(psd[1], &ret_code, 1)) != 1) {
	    diag(COMPONENT, DIAG_ERROR, 
		 "input_setup: read error %s", strerror(errno));
	    goto menu;
	}
	switch(ret_code) {
	case 0:
	    return(1);		/* we are about to quit */
	case 1:
	    goto menu;
	default:
	    diag(COMPONENT, DIAG_WARN, 
		 "user_input: got trouble when processing command (%d)", 
		 type);
	    goto menu;
	}
    default:
	diag(COMPONENT, DIAG_ERROR, 
	     "user_input: %d BAD CHOICE", choice);
	fflush(stdout);
	(void) sleep(2);
	goto menu;
    }	
}

int process_cmd(void)
{
    int ret = 0, len;
    u_int8_t type, ret_code;
    struct child_pmsg rmsg;
    enum mcl_type mtype;
    pid_t pid;
    int status;

    if ((ret = read(psd[0], &type, 1)) != 1) {
	diag(COMPONENT, DIAG_ERROR, 
	     "process_cmd: read error %s", strerror(errno));
	ret_code = 0;
	goto feedback;
    }

    switch (type) {
    case ch_dispose_clmb:
    case ch_dispose_mcs:
	len = sizeof(rmsg);
	bzero(&rmsg, len);
	rmsg.type = type;
	if ((ret = read(psd[0], (u_int8_t *)&rmsg + 1, len - 1)) != len - 1) {
	    diag(COMPONENT, DIAG_ERROR, 
		 "process_cmd: read error %s", strerror(errno));
	    ret_code = 0;
	    goto feedback;
	}
    default:
	break;
    }

    printf("\n\n");
    
    switch (type) {
	
    case ch_show_calls:		/* (1) Show Current ATM VCs */
	show_atmcalls();
        ret_code = 0;
        break;

    case ch_show_clmb:		/* (2.1) Show All Cluster Members */
        show_all_mclients(clmb_type);
	ret_code = 0;
        break;

    case ch_host_table:		/* (2.2) Show Host Map Table */
	show_map_table(clmb_type);
        ret_code = 0;
        break;
	
    case ch_dispose_clmb:	/* (2.3) Dispose of CLMB */
    case ch_dispose_mcs:	/* (3.3) Dispose of MCS */
	mtype = (type == ch_dispose_clmb) ? clmb_type : mcs_type;
	ret = dispose_of_mclient(mtype, rmsg.mcl_id);
	ret_code = ret ? 1 : 0;
	break;

    case ch_show_mcs:	/* (3.1) Show All MCS */
        show_all_mclients(mcs_type);
	ret_code = 0;
        break;

    case ch_server_table:	/* (3.2) Show Server Map Table */
	show_map_table(mcs_type);
        ret_code = 0;
        break;

    case ch_stats: 	/* (4) Show MARSD Stats */
	show_stats();
	ret_code = 0;
	break;

    case ch_close_all:	/* (5) Close All ATM VCs */
	ret = close_all(0);
	ret_code = ret ? 1 : 0;
	break;

    case ch_quit:	/* (6) Quit */
	if (AtmCalls != NULL) {
	    diag(COMPONENT, DIAG_WARN, 
		 "\nprocess_cmd: !!! STILL %d PENDING CALL(S)", NbCalls);
	    diag(COMPONENT, DIAG_WARN, 
		 "==> GO BACK TO MENU AND DE-REGISTER\n");
	    (void) sleep(2);
	    ret_code = 1;
	    break;
	}
	ret_code = 0;
	break;

    default:
	diag(COMPONENT, DIAG_ERROR, 
	     "process_cmd: Unknown type of command: %d", type);
	ret_code = 1;
	fflush(stdout);
	(void) sleep(2);
    }
    
 feedback:
    if ((ret = write(psd[0], &ret_code, 1)) < 0) {
	diag(COMPONENT, DIAG_ERROR, 
	      "process_cmd write: %s", strerror(errno));
	return(-1);
    }
    if ((type == ch_quit) && (ret_code == 0)) {
	diag(COMPONENT, DIAG_DEBUG, "\n!!! Terminating session");
	if ((pid = wait(&status)) < 0) {
	    diag(COMPONENT, DIAG_ERROR, "wait error: %s", strerror(errno));
	    return(-1);
	}
	return(1);
    }
    return(0);
}
