/* marsd_tab.h - MARS Daemon's table management */

/* Written 1999 by Mohsen.Souissi@inria.fr, INRIA */


#ifndef _MARSDTAB_H
#define _MARSDTAB_H

struct mcl_entry {
    TAILQ_ENTRY(mcl_entry) links;	/* glue list */
    u_int16_t mcl_id;			/* equals CMI for CLMB or MCSId */ 
    enum mcl_type mcl_type;		/* clmb_type or mcs_type */
    enum mcl_state mcl_state;		/* `none', `registering', `registered' 
					   or `deregistering' */
    u_int8_t *to_send;			/* pending msg (if any) to send
                                           back to cluster member */
    int send_len;

    struct svc_addr svc_addr;		/* ATM Address */
    struct in6_addr ia6;		/* IPv6 address */
    struct minmax_head minmax_head;	/* tq-list of min_max pairs (groups) 
					   the Cluster Member has joined 
					   sorted in ascending order 
					   if MCS, list of mgroups served */
    ATMCALL *p2p;			/* p2p atm connection with client */
    PARTY *party;			/* corresponding PARTY in CCVC/SCVC */
};

TAILQ_HEAD(mcl_head, mcl_entry);

struct map_entry {
    TAILQ_ENTRY(map_entry) links;	/* glue list */
    struct in6_addr ip6m;		/* IPv6 multicast address */
    struct svc_addr_head addr_head;	/* list of mapping ATM addresses */
};

TAILQ_HEAD(map_head, map_entry);


struct mcl_entry *alloc_new_mcl_id(enum mcl_type mtype);
struct mcl_entry *lookup_mcl_entry(struct svc_addr *addr, 
				   struct mcl_head *mcl_h);
struct mcl_entry *lookup_mcl_by_id(u_int16_t mcl_id, 
				   struct mcl_head *mcl_h);

struct map_entry *lookup_map_entry(struct map_head *mh,
				   struct in6_addr *ip6m, 
				   struct map_entry *start_e,
				   int *found);
struct mcl_entry *new_mcl_entry(enum mcl_type mtype,
				struct svc_addr *addr,
				struct in6_addr *ia6,
				ATMCALL *mcl_p2p);
struct map_entry *new_map_entry(struct in6_addr *ip6m, struct svc_addr *addr);
int is_pair_encompassed(struct minmax_pair *pair, struct minmax_head *mmh);
int is_pair_excluded(struct minmax_pair *pair, struct minmax_head *mmh);


int register_mars_client(enum mars_opcode opcode,
			 enum mcl_type mtype,
			 ATMCALL *mcl_p2p,
			 struct svc_addr *addr, 
			 struct in6_addr *ia6,
			 u_int8_t *mars_buf,
			 int mars_len);

int deregister_mars_client(enum mars_opcode opcode,
			   u_int8_t *mars_buf,
			   int mars_len,
			   struct mcl_entry *mcl_e);

void update_host_map_entry(struct svc_addr *addr,
			   struct map_entry *map_e,
			   enum mars_opcode op,
			   int *removed);

void remove_from_all_joined(struct mcl_entry *mcl_e);
void remove_from_server_map(struct mcl_entry *mcs_e);
void update_minmax_list(struct minmax_head *mmh,
			struct minmax_pair *pair,
			struct minmax_entry *marked_e);
void free_addr_entries(struct svc_addr_head *svc_head);
void free_map_entry(struct map_entry *map_e);
void free_mcl_entry(struct mcl_entry *mcl_e);

int dispose_of_mclient(enum mcl_type mtype, u_int16_t mcl_id);

void release_all_tables(void);

void show_all_mclients(enum mcl_type mtype);
void show_map_table(enum mcl_type mtype);
void show_server_map_table(void);
void show_stats(void);

extern struct mcl_head MCL_Table[2];
extern struct map_head Map_Table[2];

#endif
