/* to_mclient.c -  Generate outgoing MARS messages towards mars_clients */

/* Written 1999 by Mohsn.Souissi@inria.fr, INRIA */

#include <sys/param.h>
#include <sys/socket.h>
#include <sys/file.h> 
#include <sys/time.h>
#include <sys/types.h>
#include <sys/queue.h>

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>
#include <assert.h>


#include <net/if.h>
#include <net/if_atm.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netnatm/natm.h>

#include "../linux-atm.h"
#include "atm.h"
#include "atmsap.h"
#include "atmd.h"
#include "ipv6_tools.h"

#include "atmcall.h"
#include "scl_io.h"

#include "mars_msg.h"
#include "tables.h"
#include "mars_def.h"

#include "marsd_tab.h"
#include "to_mclient.h"
#include "marsd_io.h"


#define COMPONENT "TO_MCLIENT"

int send_map_list(ATMCALL *mcl_p2p,
		  struct mars_req *req_msg,
		  u_int8_t *sha,
		  u_int8_t *ssa,
		  u_int8_t *spa,
		  struct map_entry *map_e)
{
    struct mars_multi *multi_msg;
    u_int8_t *msg;
    u_int16_t msg_len, tnum, num_seq, nba, seqxy;
    u_int8_t *start;
    struct svc_addr_entry *walk, *resume_e;
    int error, len1, max_addr_list_len;
    int p, n = 0, k, r;		/* cf 5.1.1 for p, n, k meaning */
    enum mcl_type mtype;

    /* compute length of the first part, without address list */
    len1 = sizeof(struct mars_multi) + req_msg->fxh.shtl + req_msg->fxh.sstl 
	+ req_msg->spln + req_msg->tpln;
    
    /* compute maximum allowed lenght of address list per multi sequence */
    max_addr_list_len = MTU_DEFAULT - len1;
    p = max_addr_list_len/ATM_ESA_LEN;
    /* compute total number of addresses in mapping list */
    for (walk = LIST_FIRST(&map_e->addr_head); 
	 walk; 
	 walk = LIST_NEXT(walk, links)) {
	n++;
    }
    /* compute total number of multi sequences to be sent */
    k = (n/p);
    r = n - k*p;	/* remainder */
    if (r > 0)
	k++;

    resume_e = LIST_FIRST(&map_e->addr_head);

    for (num_seq = 1; num_seq <= k; num_seq++) {
	/* create and format the multi sequence to send */
	if (num_seq < k) {
	    tnum = p;
	    seqxy = num_seq & SEQ_Y;
	} else {
	    tnum = r;
	    seqxy = num_seq | LAST_SEQ;
	}
	msg_len = ATMLLC_LEN + len1 + tnum*ATM_ESA_LEN;
	msg = (u_int8_t *) malloc(msg_len);
	bzero(msg, msg_len);
	bcopy(&llc_mars_ctl, msg, ATMLLC_LEN);
	multi_msg = (struct mars_multi *) (msg + ATMLLC_LEN);
	mtype = mcl_p2p->mcl_e->mcl_type;
	format_multi_msg(req_msg, multi_msg, tnum, seqxy, 
			 (mtype == clmb_type) ? CurrCSN : CurrSSN);
	
	start = ((u_int8_t *) multi_msg) + sizeof(struct mars_multi);
	set_info(&start, &my_local_atm_addr, req_msg->fxh.shtl);
	/* ssa is NULL, sstl is zero */
	set_info(&start, spa, req_msg->spln);
	set_info(&start, &map_e->ip6m, req_msg->tpln);
	
	for (walk = resume_e, nba = 1; nba <= tnum; nba++) {
	    set_info(&start, walk->addr.prv, ATM_ESA_LEN);
	    walk = LIST_NEXT(walk, links);
	}
	resume_e = walk;
    
	/* send the multi sequence */
	diag(COMPONENT, DIAG_DEBUG, "TO %s: %s", 
	     mcl_name[mcl_p2p->mcl_e->mcl_type], mop_name[mars_multi]);
	if((error = send_msg_to_mclient(mcl_p2p->sd, msg, msg_len)) != 0) {
	    diag(COMPONENT, DIAG_ERROR, 
		 "send_map_list: failed to send sequence %d to client",
		 seqxy & SEQ_Y);
	    return(-1);
	}
	free(msg);
    }
    return(0);
}

int send_migrate(struct map_entry *map_e)
{
    struct mars_migr *migr_msg;
    u_int8_t *msg;
    u_int16_t msg_len, tnum;
    u_int8_t *start;
    struct svc_addr_entry *addr_e;
    int error, len1;

    /* compute length of the first part, without target ATM address  */
    len1 = sizeof(struct mars_migr) + ATM_ESA_LEN + 2*IN6_ADDR_LEN;
    
    addr_e = LIST_FIRST(&map_e->addr_head);

    /* create and format the migrate message to send */
    tnum = 1;

    msg_len = ATMLLC_LEN + len1 + tnum*ATM_ESA_LEN;
    msg = (u_int8_t *) malloc(msg_len);
    bzero(msg, msg_len);
    bcopy(&llc_mars_ctl, msg, ATMLLC_LEN);
    migr_msg = (struct mars_migr *) (msg + ATMLLC_LEN);
    format_migr_msg(migr_msg, tnum, CurrCSN);
	
    start = ((u_int8_t *) migr_msg) + sizeof(struct mars_migr);
    set_info(&start, &my_local_atm_addr, ATM_ESA_LEN);
    /* ssa is NULL, sstl is zero */
    set_info(&start, &my_local_ipv6_addr, IN6_ADDR_LEN);
    set_info(&start, &map_e->ip6m, IN6_ADDR_LEN);
	
    set_info(&start, addr_e->addr.prv, ATM_ESA_LEN);
    
    /* send the message over CCVC */
    diag(COMPONENT, DIAG_DEBUG, "To all CLMBs: %s", mop_name[mars_migrate]);
    if((error = send_msg_to_all(ccvc_sd, msg, msg_len)) != 0) {
	diag(COMPONENT, DIAG_ERROR, 
		 "send_migrate: failed to send MARS_MIGRATE message");
	return(-1);
    }

    CurrCSN++;
    free(msg);

    return(0);
}
