/* mcl_tab.h - MARS client (CLMB/MCS) table management */

/* Written 1998-1999 by Mohsen.Souissi@inria.fr, INRIA */


#ifndef _MCL_TAB_H
#define _MCL_TAB_H

enum resolv_state {				
    rs_none,		/* invalid */
    rs_resolv,		/* waiting for resolver (MARS) response */
    rs_valid,		/* valid */
    rs_holddown,	/* no resolve request can be sent in this state */
    rs_revalid,		/* revalidating multicast group */
    rs_migr		/* migrating Multicast VC */

};

/* common MARS client (CLMB/MCS) structure */
struct mars_client {
    enum mcl_type mcl_type;	/* clmb_type or mcs_type */
    u_int16_t cmi;		/* CMI if CLMB or zero if MCS */ 
    enum mcl_state mcl_state;	/* `none', `registering', `registered' 
				   or `deregistering' */
    u_int32_t mcl_sn;		/* mclient's image of MARS Sequence Number */
    struct svc_addr svc_addr;	/* local ATM Address */
    struct in6_addr ia6;	/* local IPv6 address as 'binary'... */
    struct svc_addr mars_svc_addr;	/* marsd ATM Address */
    struct in6_addr mars_ia6;	/* marsd IPv6 address as 'binary'... */
};

/* End-point entry. Some fields are useful only while building the p2mp VC */
struct ept_entry {
    LIST_ENTRY(ept_entry) links; /* glue list */
    struct resolv_entry *res_e;	 /* back pointer to englobing resolv_entry */
    struct svc_addr addr;	 /* ATM address */
    int tag_flag;		 /* tag this address if L_MULTI_ADD fails */
				 /* with cause = 49, 51, 37 or 41, (cf 5.1.3) */
    TIMER *conn_timer;		 /* reissue the L_MULTI_ADD (cf 5.1.3) */
    u_long new_delay;		 /* 2*previous delay (starting with random) */
    PARTY *party;		 /* pointer to the associated party (if any) */
};

LIST_HEAD(ept_head, ept_entry);

struct resolv_entry {
    LIST_ENTRY(resolv_entry) links; /* glue list */
    struct in6_addr ia6;
    u_int8_t rtm_op;		/* RTM_ADD / RTM_CHANGE / RTM_DELETE */
    enum resolv_state rstate;
    struct ept_head ept_head;	/* list of ATM addresses */
    struct ept_head new_head;	/* new list if revalidation */
    ATMCALL *mvc;		/* corresponding default p2m VC (if any) */
    struct atm_qos *qos;	/* qos corresponding to next curr_p2m to be 
				 established */
/* 
 * fields for tracking the MARS MULTI sequences within a multi-part 
 * MARS MULTI message
 */
    u_int32_t msn;		/* current msn, make sure it doesn't change */
    u_int16_t seq;   		/* current multi seq */
    char discard;		/* flag for discarding all multi seq */
    TIMER *conn_timer;		/* wait for the last multi seq */
};

LIST_HEAD(resolv_head, resolv_entry);

/* Structure for unacknowledged JOIN/LEAVE/MSERV/UNSERV requests */
struct unacked_req {
    LIST_ENTRY(unacked_req) links;	/* glue list */
    u_int16_t  op;			/* JOIN/LEAVE/MSERV/UNSERV */
    u_int16_t regist;			/* register as mcl/mcs ? */
    u_int8_t seq;			/* local significance to mcl/mcs */
    u_int16_t pnum;			/* number of pairs ( = 1) */
    struct svc_addr source_atm;
    struct minmax_pair pair;		/* NULL if registering/deregistering */
    TIMER *conn_timer;
    int retries;			/* decremented at each retrans. */
    u_int8_t *msg;			/* keep it for potential retrans. */
    u_int16_t msg_len;			/* msg length (if any) */
};

LIST_HEAD(unacked_head, unacked_req);


struct resolv_entry *lookup_ip6m(struct in6_addr *ip6m);
struct unacked_req *lookup_unacked(u_int16_t opcode, u_int16_t regist,
				   u_int8_t seq, u_int16_t pnum, 
				   struct svc_addr *source_atm, 
				   struct minmax_pair *first_minmax);


int belongs_to_block_list(struct in6_addr *a, struct minmax_head *mpairs);

struct ept_entry *new_ept_entry(struct svc_addr *rsvcaddr, 
				struct resolv_entry *res_e);
struct ept_entry *find_ept_entry(struct svc_addr *atm_addr,
                                 struct resolv_entry *res_e);
void free_ept_entries(struct ept_head *head);
void discard_ept_list(struct resolv_entry *entry);

struct resolv_entry *new_resolv_entry(struct in6_addr *ip6m,
				      enum resolv_state rstate,
				      struct atm_qos *qos);
struct resolv_entry *dup_resolv_entry(struct resolv_entry *res_e, 
				      struct atm_qos *qos);
void free_resolv_entry(struct resolv_entry *res_e);
int release_resolved_mgroup(struct in6_addr *ip6m);
void free_resolv_entries(struct resolv_head *res_head);

int update_membership(struct unacked_req *unacked, u_int16_t fl_layer3grp);

void build_drop_add_lists(struct resolv_entry *entry, 
			  struct ept_head *to_drop_head);

int build_mvc(struct resolv_entry *entry);
int add_bloc_parties(ATMCALL *p2mcall, struct resolv_entry *res_e);
int close_minmax_mvc(struct minmax_pair *pair);

int update_revalid(struct resolv_entry *entry);




void add_new_leaves(struct svc_addr *source_atm, struct minmax_head *mpairs);
void drop_leaves(struct svc_addr *source_atm, struct minmax_head *mpairs);

void revalidate_all(void);

void release_all_tables(void);

void show_membership(enum mcl_type mtype);
void show_resolved_mgroups(void);
void show_p2m_calls(void);
int close_all_p2m_calls(void);

extern const char *resolv_state_name[];

extern struct resolv_head Resolv_Table;
extern struct resolv_entry *Curr_Resolved;
extern struct minmax_head Minmax_Table;
extern struct unacked_head Unacked;
extern struct mars_client mcl;	/* MARS client (CLMB/MCS) */

#endif
