/* eXternal RESolver -- Francis.Dupont@inria.fr, September 1997 */
/*	Modified by Mohsen.Souissi@inria.fr, May 1999 */
#include <sys/param.h>
#include <sys/file.h>
#include <sys/socket.h>
#include <sys/sockio.h>
#include <sys/sysctl.h>
#include <sys/ioctl.h>
#include <sys/queue.h>

#include <net/if.h>

#ifndef if_type
#include <net/if_var.h>
#else
#define ifma_msghdr	ifa_msghdr
#define ifmam_msglen	ifam_msglen
#define ifmam_addrs	ifam_addrs
#define ifmam_index	ifam_index
#endif

#include <net/if_atm.h>
#include <net/if_dl.h>
#include <net/if_types.h>
#include <net/route.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netnatm/natm.h>

#include <netdb.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <strings.h>

#include "../linux-atm.h"
#include "atm.h"
#include "atmd.h"

#include "atmcall.h"
#include "scl_io.h"

#include "mars_msg.h"
#include "tables.h"

#include "mcl_tab.h"
#include "mcl_io.h"
#include "to_mars.h"

#define COMPONENT "RT_MSG"

int seq = 1;
struct {
	struct rt_msghdr m_rtm;
	char m_space[512];
} m_rtmsg0, m_rtmsg;
struct rt_addrinfo info;


#define dst		((struct sockaddr_in6 *)info.rti_info[RTAX_DST])
#define ifmamaddr	((struct sockaddr_in6 *)info.rti_info[RTAX_IFA])

#define ROUNDUP(a) \
	((a) > 0 ? (1 + (((a) - 1) | (sizeof(long) - 1))) : sizeof(long))
#define ADVANCE(x, n)	(x += ROUNDUP((n)->sa_len))


const char *rt_msgtypes[] = {
    "",
    "RTM_ADD",
    "RTM_DELETE",
    "RTM_CHANGE",
    "RTM_GET",
    "RTM_LOSING",
    "RTM_REDIRECT",
    "RTM_MISS",
    "RTM_LOCK",
    "RTM_OLDADD",
    "RTM_OLDDEL",
    "RTM_RESOLVE",
    "RTM_NEWADDR",
    "RTM_DELADDR",
    "RTM_IFINFO",
    "RTM_NEWMADDR",
    "RTM_DELMADDR",
    "RTM_EXPIRE",
    "RTM_RTLOST",
    "RTM_GETNEXT",
    "RTM_REACHHINT",
    0,
};


int process_routing_msg(int s)

{
    register struct rt_msghdr *rtm;
    register struct ifma_msghdr *ifmam;
    u_int8_t rtm_type;
    register char *cp;
    register int i, cc;
    struct sockaddr *sa;
    char str_name[IFNAMSIZ + 1];
    unsigned int if_index, rt_index;
    char v[64];
    struct in6_addr ia6m;
    struct resolv_entry *res_e = NULL;
    int error;

    bzero(str_name, IFNAMSIZ + 1);
    (void) snprintf(str_name, IFNAMSIZ, ENNAME, myitf);
    if_index = if_nametoindex(str_name);

    (void)bzero(&m_rtmsg0, sizeof(m_rtmsg0));
    bzero(&info, sizeof(info));
    
    cc = read(s, (char *)&m_rtmsg0, sizeof(m_rtmsg0));
    if (cc < 0) {
	diag(COMPONENT, DIAG_ERROR, "process_routing_msg: read error : %s",
	     strerror(errno));
	return(-1);
    }
    if (cc == 0) {
	diag(COMPONENT, DIAG_ERROR, "process_routing_msg: read returns 0?");
	return(-1);
    }

    rtm = &m_rtmsg0.m_rtm;
    
    if (rtm->rtm_version != RTM_VERSION) {
	diag(COMPONENT, DIAG_WARN, "process_routing_msg: bad version (%d)", 
	     rtm->rtm_version);
	return(0);
    }

    rtm_type = rtm->rtm_type;
    diag(COMPONENT, DIAG_DEBUG, 
	 "process_routing_msg : rtm_type = %s", rt_msgtypes[rtm_type]);
    
    switch (rtm_type) {

    case RTM_RESOLVE:
	
	cp = m_rtmsg0.m_space;
	rt_index = rtm->rtm_index;
	/* make sure it is for ATM interface */
	if (rt_index != if_index) {
	    /* ignore message silently*/
	    return(0);
	}
	info.rti_addrs = rtm->rtm_addrs;
	break;
	
    case RTM_NEWMADDR:
    case RTM_DELMADDR:
	
	ifmam = (struct ifma_msghdr *) &m_rtmsg0;
	cp = (caddr_t) (ifmam + 1);
	rt_index = ifmam->ifmam_index;
	/* make sure it is for ATM interface */
	if (rt_index != if_index) {
	    /* ignore message silently*/
	    return(0);
	}
    	info.rti_addrs = ifmam->ifmam_addrs;
	break;
	
    default:
	/* Just ignore message */
	return(0);
    }
    

    for (i = 0; i < RTAX_MAX; i++) {
	if ((info.rti_addrs & (1 << i)) == 0)
	    continue;
	info.rti_info[i] = sa = (struct sockaddr *)cp;
	ADVANCE(cp, sa);
    }
    if (cp > (char *)(&m_rtmsg0 + 1)) {
	diag(COMPONENT, DIAG_WARN, "process_routing_msg: read overflow\n");
	return(0);
    }

    switch (rtm_type) {

    case RTM_RESOLVE:
	
	if ((info.rti_addrs & RTA_DST) == 0) {
	    diag(COMPONENT, DIAG_WARN, "process_routing_msg: no destination");
	    return(0);
	}

	if (dst->sin6_family != AF_INET6) {
	    diag(COMPONENT, DIAG_WARN, 
		 "process_routing_msg: no IPv6 destination\n");
	    return(0);
	}

	bcopy(&dst->sin6_addr, &ia6m, sizeof(ia6m));
	if (IN6_IS_ADDR_MULTICAST(&ia6m) == 0)
	    return(0);		/* ignore message */

	diag(COMPONENT, DIAG_DEBUG,
	     "process_routing_msg: multicast addr: %s", 
	     inet_ntop(AF_INET6, &ia6m, v, 64));
	
	/* try to resolve multicast address */
	error = resolve_mgroup(&res_e, &ia6m, RTM_CHANGE);
	if (error != 0) {
	    diag(COMPONENT, DIAG_ERROR,
		 "process_routing_msg: failed to resolve entry");
	    return(-1);
	}

	break;
	
    case RTM_NEWMADDR:
    case RTM_DELMADDR:
	
	if (ifmamaddr->sin6_family != AF_INET6) {
	    diag(COMPONENT, DIAG_WARN, 
		 "process_routing_msg: no IPv6 address");
	    return(0);
	}
	bcopy(&ifmamaddr->sin6_addr, &ia6m, sizeof(ia6m));
	if (IN6_IS_ADDR_MULTICAST(&ia6m) == 0) {
	    diag(COMPONENT, DIAG_ERROR,  
		 "process_routing_msg: not a multicast address %s",
		 inet_ntop(AF_INET6, &ia6m, v, 64));
	    return(-1);
	}
	diag(COMPONENT, DIAG_DEBUG,
	     "process_routing_msg: multicast addr: %s", 
	     inet_ntop(AF_INET6, &ia6m, v, 64));
	
    
	if (rtm_type ==  RTM_NEWMADDR)
	    error = join_serve_mgroup(&ia6m, mars_join, FL_LAYER3GRP);
	else
	    error = leave_unserve_minmax(&ia6m, 
					 &ia6m, 
					 mars_leave, 
					 FL_LAYER3GRP);
  	if (error != 0) {

	    diag(COMPONENT, DIAG_ERROR,
		 "process_routing_msg: failed to initiate %s mgroup",
		 (rtm_type ==  RTM_NEWMADDR) ? "joining" : "leaving");
	    return(-1);
	}

	break;
	

    default:
	/* unlikely, already filtered */
	break;
    }

    return(0);
}
