/* multi_data.c - handling Multicast IPv6 Data Packets */

/* Written 1999 by Mohsen.Souissi@inria.fr, INRIA */

#include <sys/param.h>
#include <sys/socket.h>
#include <sys/file.h>
#include <sys/types.h>

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>

#include <net/if.h>
#include <net/if_atm.h>
#include <netinet/in.h>
#include <netinet/ip6.h>
#include <arpa/inet.h>
#include <netnatm/natm.h>

#include "atmd.h"

#include "../linux-atm.h"

#include "mars_def.h"
#include "multi_data.h"

#define COMPONENT "MULTI_DATA"

struct atmmllc llc_multi_def = LLC_MULTI_IPV6_DEF;

int check_llc_multi_ipv6(struct atmmllc *amlc)
{
    u_int16_t etype = 0;

    /* check llchdr field */
    if (bcmp(amlc->llchdr, &llc_multi_def.llchdr, ATMMLLC_HDRLEN) != 0) {
	diag(COMPONENT, DIAG_ERROR, 
	     "check_llc_multi_ipv6: bad ATM multicast LLC header");
	return(-1);
    }

    /* check pid field */
    if (bcmp(amlc->pid, &llc_multi_def.pid, ATMMLLC_PIDLEN) != 0) {
	diag(COMPONENT, DIAG_ERROR, 
	     "check_llc_multi_ipv6: bad ATM multicast LLC pid");
	return(-1);
    }

    /* check ether type field */
    etype = ATM_MLLC_TYPE(amlc);
    if (etype != ETHERTYPE_IPV6) {
	diag(COMPONENT, DIAG_ERROR, 
	     "check_llc_multi_ipv6: unsupported upper protocol 0x%04x", etype);
	return(-1);
    }
    
    return(0);
}

int recv_natm_multi_ipv6(int sd, u_int8_t **mbuff, u_int16_t *len)
{
    static long mabuffer[MAX_BUF_LEN / sizeof(long) + 1];
    u_int8_t *mbuffer = (u_int8_t *)mabuffer;
    struct atmmllc *atmmllc;
    int size, min_len;

    if ((size = read(sd, mbuffer, MAX_BUF_LEN)) < 0) {
	diag(COMPONENT, DIAG_ERROR, "receive_ipv6_multicast: %s", 
	     strerror(errno));
	return(-1);       
    }
    if (size == 0) {
	diag(COMPONENT, DIAG_ERROR, 
	     "recv_natm_multi_ipv6: natm_socket (%d) closed by peer", sd);
	return(-1);
    }

    min_len =  ATMMLLC_LEN + sizeof(struct ip6_hdr);
    if (size < min_len) {
	diag(COMPONENT, DIAG_ERROR,
	     "recv_natm_multi_ipv6: packet too short (got %d < %d)", 
	     size, min_len);
	return(-1);
    }
    
    *mbuff = mbuffer;
    *len = size;

    atmmllc = (struct atmmllc *)mbuffer;
    if (check_llc_multi_ipv6(atmmllc) != 0) {
	diag(COMPONENT, DIAG_ERROR,
	     "recv_natm_multi_ipv6: bad multicast LLC encapsulation");
	return(-1);
    }
    
    return(0);
}

