/* tables.h - MARSD/CLMB/MCS common table management */

/* Written 1998-1999 by Mohsen.Souissi@inria.fr, INRIA */


#ifndef _TABLES_H
#define _TABLES_H

#define IP6_ALL_NODES_GRP	"ff02::1"
#define IP6_SOL_NODE_GRPS	"ff02:0:0:0:0:1:ff00:0"
#define SOL_NODE_PREFIX_LEN	13
#define MIN_MULTI_ADDR		"ff00::"
#define MAX_MULTI_ADDR		"ffff:ffff:ffff:ffff:ffff:ffff:ffff:ffff"

enum mcl_type {
    clmb_type = 1, 
    mcs_type = 2,
};

enum mcl_state {
    mcl_none = 0,
    mcl_registering = 1,
    mcl_registered = 2,
    mcl_deregistering = 3
};

/* <min, max> pair of IPv6 multicast addresses */
struct minmax_pair {
    struct in6_addr min;
    struct in6_addr max;
    /* min = max for a single multicast address */
};

/* Multicast groups joined by a given interface */
struct minmax_entry {
    TAILQ_ENTRY(minmax_entry) links;	/* glue list */
    struct minmax_pair pair;
    u_int16_t fl_layer3grp;		/* keep track of layer3grp flag */
};

TAILQ_HEAD(minmax_head, minmax_entry);

struct in6_addr_entry {
    TAILQ_ENTRY(in6_addr_entry) links;	/* glue list */
    struct in6_addr ia6;
};

TAILQ_HEAD(in6_addr_head, in6_addr_entry);

struct svc_addr_entry {
    LIST_ENTRY(svc_addr_entry) links;
    struct svc_addr addr;
};

LIST_HEAD(svc_addr_head, svc_addr_entry);

extern const char *mcl_name[];
extern const char *mcl_state_name[];

struct svc_addr_entry *new_addr_entry(u_int8_t *tha);

void free_svc_addr_entries(struct svc_addr_head *head);

void free_in6_addr_entries(struct in6_addr_head *in6h);

void free_minmax_entries(struct minmax_head *mmh);

void insert_new_mgroup(struct minmax_head *mmh, 
		       struct in6_addr *ip6m, 
		       u_int16_t fl_layer3grp);

int remove_mgroup(struct minmax_head *mmh, 
		  struct in6_addr *ip6m);

#endif



