/* mcs_menu.c - interface menu for MCS */

/* Written 1999 by Mohsen.Souissi@inria.fr, INRIA */

#include <sys/types.h>
#include <sys/socket.h>
#include <net/if.h>
#include <netinet/in.h>
#include <net/route.h>
#include <netnatm/natm.h>
#include <arpa/inet.h>
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>
#include <signal.h>
#include <limits.h>
#include <sys/wait.h>
#include <sys/queue.h>

#include "../linux-atm.h"
#include "atmd.h"
#include "atm.h"
#include "ipv6_tools.h"
#include "misc_tools.h"

#include "atmcall.h"
#include "scl_io.h"

#include "mars_msg.h"
#include "tables.h"

#include "mcl_tab.h"
#include "mcl_io.h"
#include "to_mars.h"


#define COMPONENT	"MCS_MENU"

int psd[2];			/* socket pair descriptors */

enum menu_choice {
    ch_show_calls = 1, 
    ch_reg_dereg = 2, ch_reg = 21, ch_dereg = 22, 
    ch_msupport = 3, ch_show_supported = 31, ch_serve = 32, ch_unserve = 33,
    ch_p2m = 4, ch_show_p2m = 41, ch_setup = 42, ch_close = 43, 
    ch_close_all_p2m = 44, 
    ch_close_all = 5, ch_quit = 6
}; 
    

static void menu_display(void) 
{
    (void) sleep(1);
    printf("\n\n");
    printf("\t\t****************************************\n");
    printf("\t\t*             MCS Main MENU            *\n");
    printf("\t\t****************************************\n\n\n");
    printf("1)  Show ATM Calls\n");
    printf("2)  Register/Deregister with MARS\n");
    printf("3)  Multicast Group Support\n");
    printf("4)  Point-to-Multipoint ATM Call Management\n");
    printf("5)  Close All ATM Calls\n");
    printf("6)  Quit\n");
    printf("\nchoice: ");
}

static int input_registration(void)	/* register/deregister with MARS */
{
    int ret;
    int num, choice, len;
    char choice_str[BUF_LEN];
    u_int8_t type;

  sub_menu:
    printf("\n\n\n\n");
    printf("\t---------------------------------------------\n");
    printf("\t*              MCS Registration             *\n");
    printf("\t---------------------------------------------\n\n");
    printf("1) Register MCS\n");
    printf("2) Deregister Cluster Member\n");
    printf("3) Back to MENU\n");
    
    printf("\nchoice: ");
    if (fgets(choice_str, sizeof(choice_str), stdin) == NULL) {
	diag(COMPONENT, DIAG_ERROR, "I/O ERROR");
	abort();		/* something is going wrong */
    }

    if ((len = strlen(choice_str)) < 2)
	goto sub_menu;
    choice_str[len - 1] = '\0';	/* replacing '\n' */
    type = ch_reg_dereg;
    if ((num = str2num(choice_str)) < 0) {
	diag(COMPONENT, DIAG_ERROR, 
	     "input_registration: str2num: %s : not a number", choice_str);
	goto sub_menu; 
    }
    else {
	choice = num;
	switch (choice) {
	case 1:
	case 2:
	    type = 10*type + num;
	    break;

	case 3:
	    return(1);		/* will go back to menu_display */

	default:
	    diag(COMPONENT, DIAG_ERROR, 
		 "input_registration: %d BAD CHOICE", choice);
	    goto sub_menu;
	}
    }
    
    if ((ret = write(psd[1], &type, 1)) < 0) {
	diag(COMPONENT, DIAG_ERROR, 
	     "input_registration write: %s", strerror(errno));
	return(ret);
    }
    if (ret != 1) {
	diag(COMPONENT, DIAG_ERROR, 
	     "input_registration (case %d): bad write", type);
	return(ret);
    }

    return(0);
}

static int input_mgroup_support(void)	/* Multicast Group Support */
{
    int ret;
    int num, choice, len;
    char choice_str[BUF_LEN];
    u_int8_t type;
    struct child_pmsg wmsg;

  sub_menu:
    printf("\n\n\n\n");
    printf("\t---------------------------------------------\n");
    printf("\t*             Multicast Group Support        *\n");
    printf("\t---------------------------------------------\n\n");
    printf("1) Show Supported Multicast Groups\n");
    printf("2) Serve a Multicast Group\n");
    printf("3) Unserve a Multicast Group\n");
    printf("4) Back to MENU\n");
    
    printf("\nchoice: ");
    if (fgets(choice_str, sizeof(choice_str), stdin) == NULL) {
	diag(COMPONENT, DIAG_ERROR, "I/O ERROR");
	abort();		/* something is going wrong */
    }

    if ((len = strlen(choice_str)) < 2)
	goto sub_menu;
    choice_str[len - 1] = '\0';	/* replacing '\n' */
    type = ch_msupport;
    if ((num = str2num(choice_str)) < 0) {
	diag(COMPONENT, DIAG_ERROR, 
	     "input_mgroup_support: str2num: %s : not a number", choice_str);
	goto sub_menu; 
    }
    else {
	choice = num;
	switch (choice) {
	case 1:
	case 2:
	case 3:
	    type = 10*type + num;
	    break;

	case 4:
	    return(1);		/* will go back to menu_display */

	default:
	    diag(COMPONENT, DIAG_ERROR, 
		 "input_mgroup_support: %d BAD CHOICE", choice);
	    goto sub_menu;
	}
    }
    
    if ((type == ch_serve) || (type == ch_unserve)) {
	/* initialize pipe message and fill in it */
	bzero(&wmsg, sizeof(wmsg));
	wmsg.type = type;

	ret = get_mgroup_addr(&wmsg.ip6m);
	if (ret != 0)
	    goto sub_menu;
	if ((ret = write(psd[1], &wmsg, sizeof(wmsg))) < 0) {
	    diag(COMPONENT, DIAG_ERROR, 
		 "input_mgroup_support write: %s", strerror(errno));
	    return(ret);
	}
	if (ret != sizeof(wmsg)) {
	    diag(COMPONENT, DIAG_ERROR, "input_resolution : bad write");
	    return(1);
	}
    } else {
	if ((ret = write(psd[1], &type, 1)) < 0) {
	    diag(COMPONENT, DIAG_ERROR, 
		 "input_mgroup_support write: %s", strerror(errno));
	    return(ret);
	}
	if (ret != 1) {
	    diag(COMPONENT, DIAG_ERROR, 
		 "input_mgroup_support (case %d): bad write", type);
	    return(ret);
	}
    }

    return(0);
}

static int input_p2m(void)	/* p2m call management */
{
    int ret;
    int num, choice, len;
    char choice_str[BUF_LEN], vcc_str[BUF_LEN];
    u_int8_t type;
    u_long vcc;
    struct child_pmsg wmsg;

  sub_menu:
    printf("\n\n\n\n");
    printf("\t---------------------------------------------\n");
    printf("\t*  Point-to-Multipoint ATM Call Management  *\n");
    printf("\t---------------------------------------------\n\n");
    printf("1) Show p2m ATM Calls\n");
    printf("2) Close a p2m Call\n");
    printf("3) Close All p2m Calls\n");
    printf("4) Back to MENU\n");

    printf("\nchoice: ");
    if (fgets(choice_str, sizeof(choice_str), stdin) == NULL) {
	diag(COMPONENT, DIAG_ERROR, "I/O ERROR");
	abort();		/* something is going wrong */
    }

    if ((len = strlen(choice_str)) < 2)
	goto sub_menu;
    choice_str[len - 1] = '\0';	/* replacing '\n' */
    type = ch_p2m;
    if ((num = str2num(choice_str)) < 0) {
	diag(COMPONENT, DIAG_ERROR, 
	     "input_p2m: str2num: %s : not a number", choice_str);
	goto sub_menu; 
    }
    else {
	choice = num;
	switch (choice) {
	case 1:
	case 2:
	case 3:
	    type = 10*type + num;
	    break;

	case 4:
	    return(1);		/* will go back to menu_display */

	default:
	    diag(COMPONENT, DIAG_ERROR, "input_p2m: %d BAD CHOICE", choice);
	    goto sub_menu;
	}
    }
    
    if (type == ch_close) {
	/* initialize pipe message and fill in it */
	bzero(&wmsg, sizeof(wmsg));
	wmsg.type = type;

	/* capture useful data and send them */
    retry_vcc:
	printf("\n vcc : ");
	if (fgets(vcc_str, sizeof(vcc_str), stdin) == NULL) {
	    diag(COMPONENT, DIAG_ERROR, "I/O ERROR");
	    abort();		/* something is going wrong */
	}

	if ((len = strlen(vcc_str)) < 2) 
	    goto retry_vcc;
	vcc_str[len - 1] = '\0'; /* replacing '\n' */
	if ((ret = str2num(vcc_str)) < 0) {
	    diag(COMPONENT, DIAG_ERROR, 
		 "input_p2m: %s : Not a number", vcc_str);
	    fflush(stdout);
	    (void) sleep(1);
	    goto sub_menu;
	}
	vcc = (u_long) ret;
	wmsg.vcc = vcc;
	
	if ((ret = write(psd[1], &wmsg, sizeof(wmsg))) < 0) {
	    diag(COMPONENT, DIAG_ERROR, 
		 "input_p2m write: %s", strerror(errno));
	    return(ret);
	}
	if (ret != sizeof(wmsg)) {
	    diag(COMPONENT, DIAG_ERROR, "input_p2m : bad write");
	    return(1);
	}
    } else {
	if ((ret = write(psd[1], &type, 1)) < 0) {
	    diag(COMPONENT, DIAG_ERROR, 
		 "input_p2m write: %s", strerror(errno));
	    return(ret);
	}
	if (ret != 1) {
	    diag(COMPONENT, DIAG_ERROR, 
		 "input_p2m (case %d): bad write", type);
	    return(ret);
	}
    }

    return(0);
}

int user_input(void)
{
    int ret = 0, len, choice;
    char choice_str[BUF_LEN];
    u_int8_t type, ret_code;

 menu:
    menu_display();
    if (fgets(choice_str, sizeof(choice_str), stdin) == NULL) {
	diag(COMPONENT, DIAG_ERROR, "I/O ERROR");
	abort();		/* something is going wrong */
    }
    if ((len = strlen(choice_str)) < 2) 
	goto menu;
    choice_str[len - 1] = '\0';	/* replacing '\n' */
    if ((ret = str2num(choice_str)) < 0) {
	diag(COMPONENT, DIAG_ERROR, 
	     "user_input: str2num: %s : Not a number", choice_str);
	fflush(stdout);
	(void) sleep(2);
	goto menu;
    }
    choice = ret;
    type = (u_int8_t) choice;
    switch (choice) {
    case ch_show_calls:	/* (1) Show ATM Calls */
    case ch_close_all:	/* (5) Close All ATM Calls */
	if ((ret = write(psd[1], &type, 1)) < 0) { /* send only the type */
	    diag(COMPONENT, DIAG_ERROR, 
		  "user_input write: %s", strerror(errno));
	    goto menu;
	}
	if (ret != 1) {
	    diag(COMPONENT, DIAG_ERROR, 
		 "user_input (case %d): bad write", type);
	    goto menu;
	}
	if ((ret = read(psd[1], &ret_code, 1)) != 1) {
	    diag(COMPONENT, DIAG_ERROR, 
		 "user_input: read error %s", strerror(errno));
	    goto menu;
	}
	if (ret_code != 0)
	    diag(COMPONENT, DIAG_WARN, 
		 "user_input: got trouble when processing command (%d)", 
		 type);
	goto menu;
	
    case ch_reg_dereg:	/* (2) Registration/Deregistration */
	if (input_registration() != 0)
	    goto menu;
	if ((ret = read(psd[1], &ret_code, 1)) != 1) {
	    diag(COMPONENT, DIAG_ERROR, 
		 "user_input: read error %s", strerror(errno));
	    goto menu;
	}
	if (ret_code != 0)
	    diag(COMPONENT, DIAG_WARN, 
		 "user_input: got trouble when processing command (%d)", 
		 type);
	goto menu;
	
    case ch_msupport:	/* (3) Multicast Group Support */
	if (input_mgroup_support() != 0)
	    goto menu;
	if ((ret = read(psd[1], &ret_code, 1)) != 1) {
	    diag(COMPONENT, DIAG_ERROR, 
		 "user_input: read error %s", strerror(errno));
	    goto menu;
	}
	if (ret_code != 0)
	    diag(COMPONENT, DIAG_WARN, 
		 "user_input: got trouble when processing command (%d)", 
		 type);
	goto menu;
	
    case ch_p2m:	/* (4) Point-to-Multipoint Calls */
	if (input_p2m() != 0)
	    goto menu;
	if ((ret = read(psd[1], &ret_code, 1)) != 1) {
	    diag(COMPONENT, DIAG_ERROR, 
		 "user_input: read error %s", strerror(errno));
	    goto menu;
	}
	if (ret_code != 0)
	    diag(COMPONENT, DIAG_WARN, 
		 "user_input: got trouble when processing command (%d)", 
		 type);
	goto menu;
		
    case ch_quit:	/* (6) QUIT */
	if ((ret = write(psd[1], &type, 1)) < 0) {
	    diag(COMPONENT, DIAG_ERROR, 
		  "user_input write: %s", strerror(errno));
	    goto menu;
	}
	if (ret != 1) {
	    diag(COMPONENT, DIAG_ERROR, 
		 "user_input (case %d): bad write", type);
	    goto menu;
	}
	if ((ret = read(psd[1], &ret_code, 1)) != 1) {
	    diag(COMPONENT, DIAG_ERROR, 
		 "user_input: read error %s", strerror(errno));
	    goto menu;
	}
	switch(ret_code) {
	case 0:
	    return(1);		/* we are about to quit */
	case 1:
	    goto menu;
	default:
	    diag(COMPONENT, DIAG_WARN, 
		 "user_input: got trouble when processing command (%d)", 
		   type);
	    goto menu;
	}
    default:
	diag(COMPONENT, DIAG_ERROR, "user_input: %d BAD CHOICE", choice);
	fflush(stdout);
	(void) sleep(2);
	goto menu;
    }	
}

int process_cmd(void)
{
    int ret = 0, len;
    u_int8_t type, ret_code;
    struct child_pmsg rmsg;
    ATMCALL *call;
    pid_t pid;
    int status;
    char str_addr[64];


    if ((ret = read(psd[0], &type, 1)) != 1) {
	diag(COMPONENT, DIAG_ERROR, 
	     "process_cmd: read error %s", strerror(errno));
	ret_code = 0;
	goto feedback;
    }
  
    switch (type) {
    case ch_serve:
    case ch_unserve:
    case ch_close:
	len = sizeof(rmsg);
	bzero(&rmsg, len);
	rmsg.type = type;
	if ((ret = read(psd[0], (u_int8_t *)&rmsg + 1, len - 1)) != len - 1) {
	    diag(COMPONENT, DIAG_ERROR, 
		 "process_cmd: read error %s", strerror(errno));
	    ret_code = 0;
	    goto feedback;
	}
    default:
	break;
    }

    printf("\n\n");
    switch (type) {
    case ch_show_calls:	/* (1) Show Current ATM Calls */
	show_atmcalls();
        ret_code = 0;
        break;

    case ch_reg:	/* (2.1) MCS with MARS */
	ret = register_mars_client(mars_mserv);
	ret_code = ret ? 1 : 0;
	break;

    case ch_dereg:	/* (2.2) de-Register Cluster Member */
	ret = signoff_cluster();
	ret_code = ret ? 1 : 0;
	break;

    case ch_show_supported:	/* (3.1) Show Supported Multicast Groups */
	show_membership(mcs_type);
	ret_code = 0;
	break;
	
     case ch_serve: 	/* (3.2) SERVE Multicast group */
	/* is mgroup already supported ? */
	if (belongs_to_block_list(&rmsg.ip6m, &Minmax_Table)) {
	    diag(COMPONENT, DIAG_DEBUG, 
		 "process_cmd: mgroup %s is already supported by MCS. "
		 "No message sent",
		 inet_ntop(AF_INET6, &rmsg.ip6m, str_addr, 64));
	    ret_code = 0;
	    break;
	}
	ret = join_serve_mgroup(&rmsg.ip6m, mars_mserv, 0);
  	ret_code = ret ? 1 : 0;
	break;

    case ch_unserve: 	/* (3.3) UNSERVE Multicast group */
	/* is mgroup already supported ? */
	if (!belongs_to_block_list(&rmsg.ip6m, &Minmax_Table)) {
	    diag(COMPONENT, DIAG_DEBUG, 
		 "process_cmd: mgroup %s is not supported by MCS. "
		 "No message sent",
		 inet_ntop(AF_INET6, &rmsg.ip6m, str_addr, 64));
	    ret_code = 0;
	    break;
	}
	ret = leave_unserve_minmax(&rmsg.ip6m, &rmsg.ip6m, mars_unserv, 0);
  	ret_code = ret ? 1 : 0;
	break;

    case ch_show_p2m:	/* (4.1) Show p2m ATM Calls */
	show_p2m_calls();
	ret_code = 0;
	break;

    case ch_close:	/* (4.3) Close a p2m Call */
	/* check it is a p2m call */
	if ((call = atmcall_lookup(rmsg.vcc)) == NULL) {
	    diag(COMPONENT, DIAG_ERROR, 
		 "process_cmd: no matching atm call for VCC %ld", rmsg.vcc);
	    ret_code = 1;
	    break;
	}
	if (call->conf != root) {
	    diag(COMPONENT, DIAG_ERROR, 
		 "process_cmd: not a p2m call [VCC = %ld]", call->vcc);
	    ret_code = 1;
	    break;
	}
	ret = active_close_atmcall(0, call->vcc, call);
	ret_code = ret ? 1 : 0;
	break;

    case ch_close_all_p2m:	/* (4.4) Close All p2m Calls */
	ret = close_all_p2m_calls();
	ret_code = 0;
	break;

    case ch_close_all:	/* (5) Close All ATM Calls */
	ret = close_all(0);
	ret_code = ret ? 1 : 0;
	break;
    case ch_quit:	/* (6) Quit */ 
	if (AtmCalls != NULL) {
	    diag(COMPONENT, DIAG_WARN, 
		 "process_cmd: !!! STILL %d PENDING CALL(S)", NbCalls);
	    diag(COMPONENT, DIAG_WARN, "==> GO BACK TO MENU AND DE-REGISTER");
	    (void) sleep(2);
	    ret_code = 1;
	    break;
	}
	ret_code = 0;
	break;
    default:
	diag(COMPONENT, DIAG_ERROR, 
	     "process_cmd: Unknown type of command: %d", type);
	ret_code = 1;
	fflush(stdout);
	(void) sleep(2);
    }

 feedback:
    if ((ret = write(psd[0], &ret_code, 1)) < 0) {
	diag(COMPONENT, DIAG_ERROR, 
	      "process_cmd write: %s", strerror(errno));
	return(-1);
    }
    if ((type == ch_quit) && (ret_code == 0)) {
	diag(COMPONENT, DIAG_DEBUG, "\n!!! Terminating session");
	if ((pid = wait(&status)) < 0) {
	    diag(COMPONENT, DIAG_ERROR, "wait error: %s", strerror(errno));
	    return(-1);
	}
	return(1);
    }
    return(0);
}
