/* fromsig.c - Processing of incoming messages from atmsigd */

/* Written 1997-1999 by Mohsen.Souissi@inria.fr INRIA */

#include <sys/types.h>
#include <sys/socket.h>
#include <sys/queue.h>
#include <net/if.h>
#include <netinet/in.h>
#include <net/route.h>
#include <arpa/inet.h>
#include <netnatm/natm.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <assert.h>

#include "../linux-atm.h"

#include "atm.h"
#include "atmd.h"
#include "ipv6_tools.h"

#include "atmcall.h"
#include "scl_io.h"

#ifdef NDP_CLIENT

#include "ndpd.h"

#endif

#ifdef MARS_CLIENT

#include "tables.h"
#include "mars_msg.h"
#include "mcl_tab.h"
#include "mcl_io.h"
#include "to_mars.h"
#include "mcl_timeout.h"

#endif

#ifdef MARSD

#include "tables.h"
#include "mars_msg.h"
#include "marsd_tab.h"
#include "marsd_io.h"

#endif

#define COMPONENT "FROMSIG"

/* disable all constraints for the moment */
int atm_cac(struct atmsvc_msg *msg)
{
    return(0); 
}

static int process_connect_sdu(struct atmsvc_msg *msg)
{
    PARTY *party;
    ATMCALL *call;
    struct sockaddr_natm pvc = { sizeof(pvc), AF_NATM };
    int error = 0, on = 1;
    
#ifdef MARS_CLIENT
    struct resolv_entry *res_e;
    struct ept_entry *ept_e = NULL, *walk;
    int best_effort;
    enum mcl_type mtype = mcl.mcl_type;
    enum mars_opcode opcode;
    char v[64];
#endif

#ifdef MARSD
    struct mcl_entry *mcl_e;
#endif

    if ((call = atmcall_lookup(msg->vcc)) == NULL) {
	diag(COMPONENT, DIAG_WARN, 
	     "process_connect_sdu: no matching ATM call for "
	     "vcc %ld", msg->vcc);
	return(1);
    }
    if (call->state != connecting) {
	diag(COMPONENT, DIAG_WARN, 
	     "process_connect_sdu: as_connect: unexpected message");
	return(-1);
    }
    call->vpi = msg->pvc.snatm_vpi;
    call->vci = msg->pvc.snatm_vci;
    call->state = connected;
    diag(COMPONENT, DIAG_DEBUG, "%s pvc: [vpi, vci] = [%d, %d] ", 
	   as_name[msg->type], call->vpi, call->vci);

    /* connect the socket with remote end-point  */
    (void) snprintf(pvc.snatm_if, IFNAMSIZ, ENNAME, myitf);
    pvc.snatm_vpi = call->vpi;
    pvc.snatm_vci = call->vci;

    if (connect(call->sd, (struct sockaddr *) &pvc, 
		sizeof(struct sockaddr_natm)) < 0) {
	perror("connect");
	(void) close(call->sd);
	(void) active_close_atmcall(0, call->vcc, call);
	diag(COMPONENT, DIAG_ERROR, 
	     "process_connect_sdu: failed to connect natm socket");
	return(-1);
    }

#ifdef NDP_CLIENT
    if (setsockopt(call->sd, PROTO_NATMAAL5, NATM_IPMODE,
		   &on, sizeof(int)) < 0) {
	perror("setsockopt NATM_IPMODE");
	(void) close(call->sd);
	(void) active_close_atmcall(0, call->vcc, call);
	return(-1);
    }
    if (call->qos.txtp.traffic_class == ATM_UBR)
	connecting_circuit(call);
#endif

#ifdef MARS_CLIENT

    /* check if this atm call is the p2p VC with marsd */
    if ((call->conf == p2p) && 
	(atm_equal(&mcl.mars_svc_addr, &call->rsvcaddr, 0, 0) != 0)) {
	    p2p2mars = call;
	    p2p_sd = call->sd;
	

	/* if it is a CLMB, enable IP mode */
	if (mtype == clmb_type) {

	    struct in6_addr dst;

	    if (setsockopt(call->sd, PROTO_NATMAAL5, NATM_IPMODE, 
			   &on, sizeof(int)) < 0) {
		perror("setsockopt NATM_IPMODE");
		(void) close(call->sd);
		(void) active_close_atmcall(0, call->vcc, call);
		return(-1);
	    }
	    
	    /* 
	     * add route to special destinations via this call 
	     */

	    error = inet_pton(AF_INET6, IP6_ALL_NODES_GRP, &dst);
	    if (error < 0) {
		diag(COMPONENT, DIAG_ERROR,
		     "process_connect_sdu: inet_pton: %s\n", strerror(errno));
		return(-1);
	    }
	    error = add_ipatmroute(1, &dst, 0, call, RTM_ADD);
	    if (error != 0) {
		diag(COMPONENT, DIAG_ERROR, 
		     "process_connect_sdu: failed to add route to desitnation "
		     "%s",inet_ntop(AF_INET6, &dst, v, sizeof(v)));
		return(error);
	    }

	    error = inet_pton(AF_INET6, IP6_SOL_NODE_GRPS, &dst);
	    if (error < 0) {
		diag(COMPONENT, DIAG_ERROR,
		     "process_connect_sdu: inet_pton: %s\n", strerror(errno));
		return(-1);
	    }
	    error = add_ipatmroute(0, &dst, SOL_NODE_PREFIX_LEN, call, RTM_ADD);
	    if (error != 0) {
		diag(COMPONENT, DIAG_ERROR, 
		     "process_connect_sdu: failed to add route to desitnation "
		     "%s",inet_ntop(AF_INET6, &dst, v, sizeof(v)));
		return(error);
	    }

	    /* 
	     * @@@ later, do the same thing for "All Routers Address", 
	     * "All-dhcp-servers" and "All-dhcp-relays"
	     */
	 
	}
	
	/* go back to `register_mars_client' routine */
	opcode = (mtype == clmb_type) ? mars_join : mars_mserv;
	error = register_mars_client(opcode);
	if (error != 0) {
	    diag(COMPONENT, DIAG_ERROR, 
	     "process_connect_sdu: failed to register mars client");
	    return(-10);
	}
    }
#endif

    /* is it a p2mp VC */
    if (call->conf == root) {

	party = call->parties;
	if ((party == NULL)
	    || (msg->ep_ref != 0) 
	    || (party->pstate != ps_add_init)) {
	    diag(COMPONENT, DIAG_ERROR, 
		 "Received as_connect (p2m) in bad config or bad party state");
	    (void) active_close_atmcall(0, call->vcc, call);
            return(-1);
	}

#ifdef SCL_TEST
	if (setsockopt(call->sd, PROTO_NATMAAL5, NATM_IPMODE, 
		       &on, sizeof(int)) < 0) {
	    perror("setsockopt NATM_IPMODE");
	    (void) close(call->sd);
	    (void) active_close_atmcall(0, call->vcc, call);
	    return(-1);
	}
#endif

#ifdef MARS_CLIENT
	/* 
	 * enable IPMODE only for CLMB. 
	 * MCS must stay just as a 'relay' at level 2 (ATM)
	 */
	if (mtype == clmb_type) {
	    if (setsockopt(call->sd, PROTO_NATMAAL5, NATM_IPMODE, 
			   &on, sizeof(int)) < 0) {
		perror("setsockopt NATM_IPMODE");
		(void) close(call->sd);
		(void) active_close_atmcall(0, call->vcc, call);
		return(-1);
	    }
	}
#endif

        party->pstate = ps_active;
        diag(COMPONENT, DIAG_DEBUG, 
             "process_connect_sdu:  p2mp VC  built successfully!");


#ifdef MARS_CLIENT
	best_effort = (call->qos.txtp.traffic_class == ATM_UBR);
	    
	if ((res_e = call->res_e) != NULL) {
	    /* stop all potential hunging timers */
	    for (walk = LIST_FIRST(&res_e->ept_head); walk;
		 walk = LIST_NEXT(walk, links))
		STOP_TIMER(walk);
	    if (best_effort)
		res_e->rstate = rs_valid;
	    if ((ept_e = party->ept_e) != NULL)
		ept_e->party = party;

	    /* carry on building p2m call */
	    error = add_bloc_parties(call, res_e);
	    if (error != 0) {
		diag(COMPONENT, DIAG_WARN, 
		     "process_connect_sdu:  failed to initiate adding "
		     "some parties...!");
		return(error);
	    }

	    /* 
	     * If CLMB, and Best Effort, add IPv6 route to destination 
	     * (mgroup) via this mvc 
	     */
	    if ((mtype == clmb_type) && best_effort)  {
		error = add_ipatmroute(1, NULL, 0, call, res_e->rtm_op);
		if (error != 0) {
		    diag(COMPONENT, DIAG_ERROR, 
			 "process_connect_sdu: failed to add route to "
			 "desitnation %s",
			 inet_ntop(AF_INET6, &call->rin6, v, sizeof(v)));
		    return(error);
		}
	    }
	}
	
#endif
	
#ifdef MARSD
	/* 
	 * check if marsd is waiting for confirmation of "add_party".
	 * This may be the case if marsd has just initiated establishing
	 * CCVC/SCVC with this mars client as first leaf
	 * if it's the case, carry on registration of the 
	 * new mars client
	 */
	if (((mcl_e = party->mcl_e) != NULL) && 
	    (mcl_e->to_send != NULL)) {
	    if (mcl_e->mcl_type == clmb_type) {
		error = register_mars_client(mars_join, clmb_type, mcl_e->p2p, 
					     NULL, NULL, NULL, 0);
		if (error != 0) {
		    diag(COMPONENT, DIAG_ERROR, 
			 "process_connect_sdu: failed to add new cluster "
			 "member as a leaf of CCVC");
		    return(error);
		}
		
		Ccvc = call;
		ccvc_sd = Ccvc->sd;
	    } else {	/* mcs_type */
		error = register_mars_client(mars_mserv, mcs_type, mcl_e->p2p, 
					     NULL, NULL, NULL, 0);
		if (error != 0) {
		    diag(COMPONENT, DIAG_ERROR, 
			 "process_connect_sdu: failed to add new MCS "
			 "as a leaf of SCVC");
		    return(error);
		}

		Scvc = call;
		scvc_sd = Scvc->sd;
	    }

	    return(0);
	}
#endif
    }
	return(error);
}

static int process_indicate_sdu(struct atmsvc_msg *msg)
{
    ATMCALL *newcall;
    CONN_CONF conf;
    short ep_ref;
    PARTY *newparty = NULL;
    struct sockaddr_natm pvc = { sizeof(pvc), AF_NATM };
    int error, sd, on = 1;

    ep_ref = msg->ep_ref;
    conf = (ep_ref >= 0) ? leaf : p2p;
    if ((error = atm_cac(msg)) != 0) {
	/* CAC didn't accept the new call */
	diag(COMPONENT, DIAG_WARN, "process_indicate_sdu: call rejected");
	(void) send_atmsigd(msg->listen_vcc, msg->vcc, ep_ref, as_reject, 
			    0, NULL, NULL, NULL, NULL);
	return(1);
    }

    newcall = new_atmcall(msg->listen_vcc, msg->pvc.snatm_vpi, 
			  msg->pvc.snatm_vci, conf, 0, accepting, 
			  &msg->svc, NULL, NULL, &msg->qos, NULL);
    if (newcall == NULL) {
	diag(COMPONENT, DIAG_ERROR, 
	     "process_indicate_sdu: failed to allocate a new atmcall");
	(void) send_atmsigd(msg->listen_vcc, msg->vcc, ep_ref, as_reject, 
			    0, NULL, NULL, NULL, NULL);
	return(-1);
    }
    if (conf == leaf) {
	diag(COMPONENT, DIAG_DEBUG,
	     "Incoming p2m setup %s: %s\n with pvc: [vpi, vci] = [%d, %d] "
	       ", <EPR = %d>", (ep_ref > 0) ? "(actually add_party)" : " ",  
	       as_name[msg->type], newcall->vpi, newcall->vci, ep_ref);
	newparty = new_party(newcall, (u_int16_t) ep_ref, ps_add_recv, 
			     &msg->svc, NULL, NULL, NULL);
	if (newparty == NULL) {
	    diag(COMPONENT, DIAG_ERROR, 
		 "process_indicate_sdu: failed to allocate a new party");
	    (void) send_atmsigd(msg->listen_vcc, msg->vcc, ep_ref, as_reject, 
				0, NULL, NULL, NULL, NULL);
	    goto teardown;
	}
	newcall->nbactp++;
    } else
	diag(COMPONENT, DIAG_DEBUG,
	     "Incoming p2p setup %s pvc: [vpi, vci] = [%d, %d]", 
	       as_name[msg->type], newcall->vpi, newcall->vci);
    /* creating a NATM socket for the new ATM call */
    if ((sd = new_natm_sock(newcall)) < 0) {
	diag(COMPONENT, DIAG_ERROR, 
	     "process_indicate_sdu: failed to create a NATM socket");
	(void) send_atmsigd(msg->listen_vcc, msg->vcc, ep_ref, as_reject, 
			    0, NULL, NULL, NULL, NULL);
	goto teardown;
    }
    newcall->sd = sd;

    /* connect the socket with remote end-point  */
    (void) snprintf(pvc.snatm_if, IFNAMSIZ, ENNAME, myitf);
    pvc.snatm_vpi = newcall->vpi;
    pvc.snatm_vci = newcall->vci;

    if (connect(newcall->sd, (struct sockaddr *) &pvc, 
		sizeof(struct sockaddr_natm)) < 0) {
	perror("connect");
	(void) close(newcall->sd);
	(void) send_atmsigd(msg->listen_vcc, msg->vcc, ep_ref, as_reject, 
			    0, NULL, NULL, NULL, NULL);
	diag(COMPONENT, DIAG_ERROR, 
	     "process_indicate_sdu: failed to connect natm socket");
	goto teardown;
    }

#if defined(NDP_CLIENT) || defined(SCL_TEST)

    if (setsockopt(newcall->sd, PROTO_NATMAAL5, NATM_IPMODE, 
		   &on, sizeof(int)) < 0) {
	perror("setsockopt NATM_IPMODE");
	(void) close(newcall->sd);
	(void) send_atmsigd(msg->listen_vcc, msg->vcc, ep_ref, as_reject,
			    0, NULL, NULL, NULL, NULL);
	diag(COMPONENT, DIAG_ERROR,
	     "process_indicate_sdu: failed to force IP mode");
	goto teardown;
    }

#endif

    error = send_atmsigd(msg->vcc, msg->listen_vcc, ep_ref, as_accept, 
		       0, NULL, NULL, NULL, NULL);
    if (error != 0) {
	diag(COMPONENT, DIAG_ERROR,
	     "process_indicate_sdu: failed to send as_accept");
	(void) send_atmsigd(msg->listen_vcc, msg->vcc, ep_ref, as_reject, 
			    0, NULL, NULL, NULL, NULL);
	goto teardown;
    }
    append_atmcall(newcall);
    if (CurrVcc < newcall->vcc)
	CurrVcc = newcall->vcc;
    NbCalls++;
    return(0);
 teardown:
    if (newparty != NULL)
	free(newparty);
    free(newcall);
    return(-1);
}


static int process_accept_sdu(struct atmsvc_msg *msg)
{
    ATMCALL *call;
    PARTY *party;
    int on = 1; 
  
    if ((call = atmcall_lookup(msg->vcc)) == NULL) {
	diag(COMPONENT, DIAG_WARN, 
	     "process_accept_sdu: no matching ATM call for vcc %ld", 
	     msg->vcc);
	return(1);
    }
    if (call->state != accepting) {
	diag(COMPONENT, DIAG_WARN, 
	     "process_accept_sdu: as_accept: unexpected message");
	return(1);
    }
    call->state = connected;

    if (call->conf == p2p) {
	diag(COMPONENT, DIAG_DEBUG, 
	     "a new p2p ATM call accepted with vcc %ld", msg->vcc);
#ifdef NDP_CLIENT
	if (call->qos.txtp.traffic_class == ATM_UBR)
	    connecting_circuit(call);
#endif
	return(0);
    } /* else it must be in 'leaf' conf */

    if (call->conf != leaf) {
	/* this is unlikely to happen (paranoid) */
	diag(COMPONENT, DIAG_ERROR, 
	     "process_accept_sdu: invalid 'accept' message (in p2m conf)");
	goto teardown;
    }
    
    party = call->parties;
    if ((party == NULL) || (party->pstate != ps_add_recv)) {
	diag(COMPONENT, DIAG_ERROR, 
	     "process_accept_sdu: (p2m) as_accept in bad conf or"
	     " bad party state");
	goto teardown;
    }
    diag(COMPONENT, DIAG_DEBUG, 
	 "a new ATM call leaf accepted with <vcc =%ld>", msg->vcc);
    party->pstate = ps_active;
    
#ifdef MARS_CLIENT
	
    /* 
     * if and ADD_PARTY was initiated by marsd, 
     * mark this call as CVC (CCVC/SCVC from the client's viewpoint)
     */
    
    if (atm_equal(&mcl.mars_svc_addr, &call->rsvcaddr, 0, 0)) {
	Cvc = call;
	cvc_sd = call->sd;

    }
 
    /* 
     * enable NATM_IPMODE for that leaf if CLMB 
     * This leaf may also be a party of a real multicast VC 
     */

    if (mcl.mcl_type == clmb_type)
	if (setsockopt(call->sd, PROTO_NATMAAL5, NATM_IPMODE, 
		       &on, sizeof(int)) < 0) {
	    perror("setsockopt NATM_IPMODE");
	    goto teardown;
	}
	
#endif
    
    return(0);

 teardown:
    (void) close(call->sd);
    (void) active_close_atmcall(0, call->vcc, call);
    return(-1);
}

static int process_add_sdu(struct atmsvc_msg *msg)
{
    PARTY *party;

    if ((party = party_lookup(msg->vcc, msg->ep_ref)) == NULL) {
	diag(COMPONENT, DIAG_WARN, 
	     "process_add_sdu: no matching party for "
	     "[Vcc = %ld, EPR = %d]", msg->vcc, msg->ep_ref);
	return(1);
    }
    if ((party->call->state != connected) || (party->pstate != ps_add_init)) {
	diag(COMPONENT, DIAG_WARN, 
	     "Received as_add in bad call/party states <%s, %s>",
	       CallStateName[party->call->state], 
	       PartyStateName[party->pstate]);
	return(1);		/* ignore the message */
    }

#ifdef MARS_CLIENT
    {
	struct ept_entry *ept_e;
	
	ept_e = party->ept_e;
        if (ept_e != NULL) {
	    /* link it with its corresponding party structure */
	    ept_e->party = party;
        }
    }
#endif
    
    diag(COMPONENT, DIAG_DEBUG, 
	 "process_add_sdu: add party succeeded [Vcc = %ld, EPR = %d]",
	 party->call->vcc, party->ep_ref);
    party->pstate = ps_active;
    
#ifdef MARSD
    {
	/* 
	 * check if marsd is waiting for confirmation of "add_party".
	 * if it's the case, carry on the registration of the new mars client
	 */
	struct mcl_entry *mcl_e;
	enum mars_opcode opcode;
	int error;
	
	if (((mcl_e = party->mcl_e) != NULL) 
	    && (mcl_e->to_send != NULL)) {
	    if (mcl_e->p2p == NULL) {
		diag(COMPONENT, DIAG_ERROR, 
		     "process_add_sdu: no p2p VC available with client");
		return(-1);
	    }
	    opcode = (mcl_e->mcl_type == clmb_type) 
		? mars_join : mars_mserv;
	    error = register_mars_client(opcode, mcl_e->mcl_type, mcl_e->p2p, 
					    NULL, NULL, NULL, 0);
	    if (error != 0) {
		diag(COMPONENT, DIAG_ERROR, 
		     "process_add_sdu: failed to add new mars client "
		     "as a leaf");
		return(1);
	    }
	}
    }
#endif
    
    return(0);
}

static int process_error_sdu(struct atmsvc_msg *msg)
{
    ATMCALL *call = NULL;
    PARTY *party = NULL;
    
    if (msg->vcc == 0) {
	diag(COMPONENT, DIAG_ERROR, 
	     "process_error_sdu: failed to register with atmsigd");
	return(-10);
    }
    /* is it an error on a party ? */
    if (msg->ep_ref >= 0) {
	if ((party = party_lookup(msg->vcc, msg->ep_ref)) == NULL) {
	    diag(COMPONENT, DIAG_WARN, 
		 "process_error_sdu: no matching party for "
		 "[Vcc = %ld, EPR = %d]", msg->vcc, msg->ep_ref);
	    return(1);
	}
#ifdef MARS_CLIENT
        {
            struct ept_entry *ept_e;
            int cause_code, error;
            u_long delay;
            
            if ((ept_e = party->ept_e) != NULL) {
		/* see 5.1.3 */
                cause_code = - msg->reply;
                diag(COMPONENT, DIAG_DEBUG, 
		     "process_error_sdu: received error %d for party "
		     "[Vcc = %ld, EPR = %d]", cause_code, msg->vcc,
                       msg->ep_ref);
                switch(cause_code) {
		case 37:	/* user cell rate not available - UNI 3.1 */
		case 41:	/* temporary failure */
		case 35:	/* 
				 * requested VPI/VCI not available 
				 * not in RFC but may be generated by Cellplex
				 * (3Com) it can then be considered like '41' 
				 */
		case 49:	/* QoS unavailable */
		case 51:	/* user cell rate not available - UNI 3.0 */
		    /* tag the end point */
		    ept_e->tag_flag = 1;
		    /* 
		     * if the p2mp VC isn't established yet, retry 
		     * atmcall_setup 
		     */
		    if (msg->ep_ref == 0) {
			/* 
			 * if it's not the first time the ept has been tried 
			 * (i.e. ept_e->new_dela != 0), delay retrying 
			 * that ept 
			 */
			if (ept_e->new_delay != 0) {
			    STOP_TIMER(ept_e);
			    ept_e->new_delay *= 2;
			    start_timer(ept_e->new_delay, 
					on_RANDOM_FIRST_EPT, 
					ept_e);
			    diag(COMPONENT, DIAG_DEBUG, 
				 "process_error_sdu: doubling "
				 "the timer delay...");
			    return(0);
			}
			/* it's the first time we try this ept entry */
			/* cf 5.1.3 */
			error = build_mvc(ept_e->res_e);
			return(error);
		    }
		    /* schedule reiussueing ADD_MULTI */
		    if (ept_e->new_delay != 0)
			ept_e->new_delay *= 2;
		    else
			ept_e->new_delay = get_random_delay(5*T_SEC, 10*T_SEC);
		    start_timer(ept_e->new_delay, on_RANDOM_EPT, ept_e);
		    free_party(party);
		    return(0);
		default:
		    diag(COMPONENT, DIAG_DEBUG, 
			 "process_error_sdu: freeing associated endpoint "
			 "entry...");
		    diag(COMPONENT, DIAG_ERROR, 
			 "process_error_sdu: freeing party...");
		    STOP_TIMER(ept_e);
		    LIST_REMOVE(ept_e, links);
		    free(ept_e);
		    party->ept_e = NULL;
		}
            }
        }
#endif        

	/* 
	 * @@@@ later... handle cause '35' in the general case 
	 * should retry adding every failing party because of '35'
	 */

#ifdef MARSD
	{
	    /*
	     * before releasing the party structure,  
	     * force MARS to remove any reference to MARS Client in Map 
	     * Tables, this is paranoid since MARS Client is supposed to have 
	     * left/unserved all joined/served groups. Afterwards, release 
	     * corresponding mars client (if any)
	     */
	    
	    struct mcl_entry *mcl_e;
	    
	    if ((mcl_e = party->mcl_e) != NULL) {
		mcl_e->p2p->mcl_e = NULL;
		party->mcl_e = NULL;
		if (mcl_e->mcl_type == clmb_type)
		    remove_from_all_joined(mcl_e);
		else
		    remove_from_server_map(mcl_e);
		TAILQ_REMOVE(&MCL_Table[mcl_e->mcl_type - 1], mcl_e, links);
		free_mcl_entry(mcl_e);
	    }
	}
#endif
	free_party(party);
	return(0);
    }
    
    if ((call = atmcall_lookup(msg->vcc)) == NULL) {
	diag(COMPONENT, DIAG_WARN, 
	     "process_error_sdu: no matching ATM call for vcc %ld", 
	     msg->vcc);
	return(1);
    }

#ifdef MARSD
    /* if it is CCVC/SCVC, reset it */
    if (call == Ccvc)
	Ccvc = NULL;

    if (call == Scvc)
	Scvc = NULL;
#endif

    free_atmcall(call);
    NbCalls--;
    return(0);
}

static int process_drop_sdu(struct atmsvc_msg *msg)
{
    PARTY *party;
    
    if ((party = party_lookup(msg->vcc, msg->ep_ref)) == NULL) {
	diag(COMPONENT, DIAG_WARN, 
	     "process_drop_sdu: no matching party for "
	     "[Vcc = %ld, EPR = %d]", msg->vcc, msg->ep_ref);
	return(1);
    }
    switch (party->pstate) {
        case ps_drop_init:	/* active drop confirm. */
            diag(COMPONENT, DIAG_DEBUG, 
		 "process_drop_sdu: dropping party [Vcc = %ld, EPR = %d] "
		 "confirmed", 
	       party->call->vcc, party->ep_ref);
            break;
        case ps_active:		/* passive_drop */
            diag(COMPONENT, DIAG_DEBUG, 
		 "passive dropping party [Vcc = %ld, EPR = %d]", 
		 party->call->vcc, party->ep_ref);
            break;
        case ps_add_init:
            diag(COMPONENT, DIAG_WARN, 
		 "process_add_sdu: add party rejected by network");
            break;
        default:
            diag(COMPONENT, DIAG_WARN, 
		 "process_drop_sdu: received as_drop in bad "
		 "party state %s ", PartyStateName[party->pstate]);
    }
    
#ifdef MARS_CLIENT
    {
        struct ept_entry *ept_e;
        
            /* check for associated ept entry */
        if ((ept_e = party->ept_e) != NULL) {
            diag(COMPONENT, DIAG_DEBUG, 
		 "process_drop_sdu: freeing corresponding endpoint entry...");
	    STOP_TIMER(ept_e);
            LIST_REMOVE(ept_e, links);
            free(ept_e);
            party->ept_e = NULL;
        }
    }
#endif


#ifdef MARSD
    {

	struct mcl_entry *mcl_e;
	int error;
	enum mars_opcode opcode;
	
	/* get corresponding mars client entry (if any) */
	mcl_e = party->mcl_e;
	

	if (mcl_e != NULL) {
	    /* 
	     * check if the "DROP_PARTY" has been received unexpectedly 
	     * (e.g. mars client crash). If so, release all related 
	     * structures while dropping the corresponding party
	     */

	    if ((mcl_e->mcl_state == mcl_registered)) {
		
		if (mcl_e->mcl_type == clmb_type)
		    remove_from_all_joined(mcl_e);
		else
		    remove_from_server_map(mcl_e);
		
		/* release mars client entry */
		TAILQ_REMOVE(&MCL_Table[mcl_e->mcl_type - 1], mcl_e, links);
		diag(COMPONENT, DIAG_DEBUG, 
		     "process_drop_sdu: de-registered %s with "
		     "[MCLId = %d]", mcl_e->mcl_type, mcl_e->mcl_id);
		free_mcl_entry(mcl_e);
	    } else {

		/* 
		 * check if marsd is waiting for confirmation of "DROP_PARTY"
		 * if it's the case, carry on the de-registration of the 
		 * corresponding mars client
		 */

		if ((mcl_e->mcl_state == mcl_deregistering) 
		    && (mcl_e->to_send != NULL)) {
		    
		    if (mcl_e->p2p == NULL) {
			diag(COMPONENT, DIAG_ERROR, 
			     "process_drop_sdu: no p2p VC available "
			     "with mclient");
			return(-1);
		    }
		    opcode = (mcl_e->mcl_type == clmb_type) 
			? mars_leave : mars_unserv;

		    error = deregister_mars_client(opcode, NULL, 0, mcl_e);
		    if (error != 0) {
			diag(COMPONENT, DIAG_ERROR, 
			     "process_drop_sdu: failed to drop mars client "
			     "as a leaf");
			return(error);
		    }
		}
	    }
	}
    }
#endif

    free_party(party);
    return(0);
}

static int process_close_sdu(struct atmsvc_msg *msg)
{
    ATMCALL *call;

  
    if ((call = atmcall_lookup(msg->vcc)) == NULL) {
	diag(COMPONENT, DIAG_WARN, 
	     "process_close_sdu: No such call "
	     "<Vcc =%ld>, message ignored", msg->vcc);
	return(1);
    }

    switch (call->state) {
    case disconnecting:
	/* active close confirm. */
	diag(COMPONENT, DIAG_DEBUG, 
	     "closing atm call <Vcc = %ld> confirmed", call->vcc);
	goto teardown;
    case connected:
	/* passive close for p2p/leaf conf, or confirm. of last active drop */
	diag(COMPONENT, DIAG_DEBUG, 
	     "process_close_sdu: PASSIVE CLOSE, <Vcc = %ld>, <reply = %d>", 
	     msg->vcc, msg->reply);
	break;
    case connecting:
	diag(COMPONENT, DIAG_WARN, 
	     "process_close_sdu: connect rejected by network");
	break;
    default:
	diag(COMPONENT, DIAG_WARN, 
	     "process_close_sdu: received CLOSE in bad "
	     "state: %s", CallStateName[call->state]);
    }
    
#ifdef MARS_CLIENT
    {
	/* 
	 * This might be a Multicast VC that has just been torn down 
	 * (ie, last party has just been dropped)
	 * In that case, release the corresponding resolve_entry 
	 */
	
	struct resolv_entry *res_e;
	
	if ((res_e = call->res_e) != NULL) {
	    STOP_TIMER(res_e);
	    /* the default resolve entry is chained, cut links */
	    if (res_e->qos == NULL)	
		LIST_REMOVE(res_e, links);
	    res_e->mvc = NULL;
	    free(res_e);
	    /* delete route if any */
	    if (call->rt_count != 0)
		(void) delete_ipatmroute(call->vcc, 1, NULL, 0, call);
	}
    }
#endif

#ifdef MARSD
    {
	if ((call == Ccvc) || (call == Scvc)) {
	    
	    struct mcl_entry *mcl_e;
	    PARTY *party;
	    int error;
		
	    /* get the last remaining party, it must be non NULL */
	    party = call->parties;
	    /* get corresponding mars client entry (if any) */
	    mcl_e = party->mcl_e;
	    
	    if (mcl_e != NULL) {
		/* 
		 * check if the "CLOSE" has been received unexpectedly 
		 * (e.g. last mars client crash). If so, release all related 
		 * structures while dropping the corresponding party
		 */
		
		if ((mcl_e->mcl_state == mcl_registered)) {
		    
		    if (mcl_e->mcl_type == clmb_type)
			remove_from_all_joined(mcl_e);
		    else 
			remove_from_server_map(mcl_e);
			
		    /* release mars client entry */
		    TAILQ_REMOVE(&MCL_Table[mcl_e->mcl_type - 1], mcl_e, links);
		    diag(COMPONENT, DIAG_DEBUG, 
			 "process_drop_sdu: de-registered %s with "
			 "[MCLId = %d]", mcl_e->mcl_type, mcl_e->mcl_id);
		    free_mcl_entry(mcl_e);
		} else {
		    
		    /* 
		     * check if marsd is waiting for confirmation of 
		     * "drop_party". Actually, the confirm. of the last 
		     * DROP_PARTY is received as CLOSE. If it's the case, 
		     * carry on the de-registration of the corresponding 
		     * mars client
		     */
		    
		    if ((mcl_e->mcl_state == mcl_deregistering)
			&& (mcl_e->to_send != NULL)) {
			
			if (mcl_e->p2p == NULL) {
			    diag(COMPONENT, DIAG_ERROR, 
				 "process_close_sdu: no p2p VC available "
				 "with mclient");
			    return(-1);
			}
			
			if (mcl_e->mcl_type == clmb_type) {
			    error = deregister_mars_client(mars_leave, 
							   NULL, 0, mcl_e);
			    if (error != 0) {
				diag(COMPONENT, DIAG_ERROR, 
				     "process_close_sdu: failed to close CCVC");
				return(error);
			    }
			} else {
			    error = deregister_mars_client(mars_unserv, 
							   NULL, 0, mcl_e);
			    if (error != 0) {
				diag(COMPONENT, DIAG_ERROR, 
				     "process_close_sdu: failed to close SCVC");
				return(error);
			    }
			}
		    }
		}
		
	    } /* if (mcl_e != NULL) */
	    
	    if (call == Ccvc) {
		/* set CCVC to NULL */
		Ccvc = NULL;
		ccvc_sd = -1;
	    } else {
		/* set SCVC to NULL */
		Scvc = NULL;
		scvc_sd = -1;
	    }
	}
    }
    
#endif


    (void) send_atmsigd(call->vcc, 0, -1, as_close, 
			0, NULL, NULL, NULL, NULL);
    
 teardown:
    /* delete ipatm route if any */
#ifndef NDP_CLIENT
#ifndef MARS_CLIENT
    if (call->route != 0)
	(void) delete_ipatmroute(call->vcc, 1, NULL, 0, call);
#else
    if (call->rt_count != 0)
	(void) delete_ipatmroute(call->vcc, 1, NULL, 0, call);
#endif
#else
    if (call->ndp != NULL)
	(void) delete_ipatmroute(call->vcc, 1, NULL, 0, call);
#endif

    /* close socket with peer if any */
    if (call->sd > 0)
        (void) close(call->sd);

#ifdef MARS_CLIENT
    {
	/* 
	 * This might be a Multicast VC waiting for close confirm
	 */
	
	struct resolv_entry *res_e;
	int error;

	if ((res_e = call->res_e) != NULL) {
	    STOP_TIMER(res_e);
	    res_e->mvc = NULL;
	    /* if it mars client is migrating a multicast VC */
	    if (res_e->rstate == rs_migr) { /* it must be a CLMB */
		/* build the new p2mp VC */
		error = migrate_mvc(res_e);
		if (error != 0) {
		    diag(COMPONENT, DIAG_ERROR,
			 "process_close_sdu: failed to initiate building "
			 "the new p2mp VC");
		    return(-1);
		}
	    } else {
		/* release resolve entry */
		if (res_e->qos == NULL)
		    LIST_REMOVE(res_e, links);
		free_resolv_entry(res_e);
	    }
	}
	
	if (call == p2p2mars) {
	    /* if it is a CLMB delete routes via p2p2mars  */
	    if (mcl.mcl_type == clmb_type) {

		struct in6_addr dst;
		
		error = inet_pton(AF_INET6, IP6_ALL_NODES_GRP, &dst);
		if (error < 0) {
		    diag(COMPONENT, DIAG_ERROR,
			 "process_connect_sdu: inet_pton: %s\n", 
			 strerror(errno));
		    return(-1);
		}
		(void) delete_ipatmroute(0, 1, &dst, 0, call);

		error = inet_pton(AF_INET6, IP6_SOL_NODE_GRPS, &dst);
		if (error < 0) {
		    diag(COMPONENT, DIAG_ERROR,
			 "process_connect_sdu: inet_pton: %s\n", 
			 strerror(errno));
		    return(-1);
		}

		(void) delete_ipatmroute(0, 0, &dst, 
					 SOL_NODE_PREFIX_LEN, call);
		
		/* 
		 * @@@ later, do the same thing for "All Routers Address", 
		 * "All-dhcp-servers" and "All-dhcp-relays"
		 */
		
	    }
	    p2p2mars = NULL;
	    p2p_sd = -1;
	}

	/* if it is Control VC (CCVC/SCVC) */
	if (call == Cvc) {
	    Cvc = NULL;
	    cvc_sd = -1;
	    /* mars client has been dropped unexpectedly */
	    if (mcl.mcl_state == mcl_registered)
		if (close_all(0) != 0)
		    diag(COMPONENT, DIAG_WARN, 
			 "failed to close all ATM calls");
	}
    }
#endif

#ifdef MARSD
    {
	
	struct mcl_entry *mcl_e;
	enum mcl_type mtype;
	
	
	/* if it is a CLOSE confirmation for CCVC or SCVC */
	if (call == Ccvc) {
	    /* set CCVC to NULL */
	    Ccvc = NULL;
	    ccvc_sd = -1;
	} 
	
	if (call == Scvc) {
	    /* set SCVC to NULL */
	    Scvc = NULL;
	    scvc_sd = -1;
	}
	
	/*
	 * if the de-registering mars client initiated closing its 
	 * p2p VC with marsd, marsd will close that p2p VC and 
	 * release the mars client entry
	 */
	
	if (((mcl_e = call->mcl_e) != NULL) 
	    && (mcl_e->mcl_state == mcl_deregistering)) {
	    
	    if (mcl_e->p2p != NULL) {
		mcl_e->p2p->mcl_e = NULL;
		mcl_e->p2p = NULL;
	    }
	    
	    mtype = mcl_e->mcl_type;

	    /* 
	     * force removing any reference to MARS Client in Map Tables 
	     * This is paranoid for MARS Client is supposed to have 
	     * left/unserved all joined/served groups
	     */
	    if (mtype == clmb_type)
		remove_from_all_joined(mcl_e);
	    else
		remove_from_server_map(mcl_e);
	    

	    /* release mars client entry */
	    TAILQ_REMOVE(&MCL_Table[mtype -1], mcl_e, links);
	    diag(COMPONENT, DIAG_DEBUG, 
		 "process_close_sdu: de-registered %s (MCLId = %d)", 
		 mcl_name[mtype], mcl_e->mcl_id);
		
	    /* free mars client entry */
	    free_mcl_entry(mcl_e);
	}
    }
    
#endif

    free_atmcall(call);
    NbCalls--;
#if defined(MARS_CLIENT) || defined(MARSD)
    if (NbCalls == 0) {
	/* force cleaning all internal tables */
	release_all_tables();
    }
#endif
    return(0);
}


int dispatch(struct atmsvc_msg *msg, enum atmsvc_msg_type *type)
{
    int ret;
    
    *type = msg->type;
    switch (msg->type) {
    case as_okay:
	if (msg->vcc != 0)	/* answer for "as_bind" or "as_listen" */ 
	    diag(COMPONENT, DIAG_WARN, 
		 "Received OK: unexpected message");
	return(0);
    case as_connect:		/* connect_confirm sdu */
	ret = process_connect_sdu(msg);
	return(ret);
    case as_indicate:
	ret = process_indicate_sdu(msg);
	return(ret);
    case as_accept:		/* accept_confirm sdu */
	ret = process_accept_sdu(msg);
	return(ret);
    case as_add:		/* add_confirm sdu */
	ret = process_add_sdu(msg);
	return(ret);
    case as_error: 
	diag(COMPONENT, DIAG_ERROR, 
	     "ERROR RECEIVED, Clearing end point");
	ret = process_error_sdu(msg);
	return(ret);
	/* passive drop or drop_confirm sdu if requested by client */
    case as_drop:
	ret = process_drop_sdu(msg);
	return(ret);
	/* passive close or close_confirm sdu if requested by client */
    case as_close:
	ret = process_close_sdu(msg);
	return(ret);
    default:
	diag(COMPONENT, DIAG_WARN, 
	     "invalid message %d, message ignored", 
	       (int) msg->type);
	return(1);
    }
}


int from_atmsigd(struct atmsvc_msg *msg, enum atmsvc_msg_type *type)
{
    int ret;

    diag(COMPONENT, DIAG_DEBUG, "\nFROM ATMSIGD  <<<----------------<<<");
    diag(COMPONENT, DIAG_DEBUG, 
	 "\t%s <vcc %ld, listen_vcc %ld>,  <EPR = %d>", 
	 as_name[msg->type], msg->vcc, msg->listen_vcc, msg->ep_ref);
    ret = dispatch(msg, type);
    return(ret);
}
