/* ndp_menu.c - Interface MENU NDPD */

/* Written 1998-1999 by Mohsen.Souissi@inria.fr, INRIA */

#include <sys/types.h>
#include <sys/socket.h>
#include <sys/wait.h>
#include <net/if.h>
#include <netinet/in.h>
#include <netnatm/natm.h>
#include <arpa/inet.h>
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <signal.h>
#include <limits.h>
#include <netdb.h>

#include "../linux-atm.h"
#include "atmd.h"
#include "atm.h"
#include "misc_tools.h"
#include "ipv6_tools.h"

#include "atmcall.h"
#include "scl_io.h"

#include "ndpd.h"

#define COMPONENT	"NDP_MENU"

int psd[2];			/* socket pair descriptors */

enum menu_choice {
    ch_show = 1, ch_setup = 2, 
    ch_close = 3, ch_close_all = 4, ch_quit = 5
}; 
    
static void menu_display(void) 
{
    (void) sleep(1);
    printf("\n\n");
    printf("\t\t****************************************\n");
    printf("\t\t*           NDP Main Menu              *\n");
    printf("\t\t****************************************\n\n\n");
    printf("1) SHOW Current ATM Calls\n");
    printf("2) SETUP a New Point to Point ATM Call\n");
    printf("3) CLOSE an ATM Call (deletes route if any)\n");
    printf("4) CLOSE ALL \n");
    printf("5) QUIT\n");
    printf("\nchoice: ");
}

static int input_setup(void)		/* Setup a new ATM Call */
{
    int fw_traffic_class, bw_traffic_class, max_fw_pcr, max_bw_pcr;
    int max_fw_sdu, max_bw_sdu;
    int ret, choice, len;
    char choice_str[BUF_LEN], remote_name[NAME_LEN];
    u_int8_t type = 2;
    struct child_pmsg wmsg;

 retry_host:
    printf("\n Remote Host Name: ");
    if (fgets(remote_name, sizeof(remote_name), stdin) == NULL) {
	diag(COMPONENT, DIAG_ERROR, "I/O ERROR");
	abort();		/* something is going wrong */
    }



    remote_name[strlen(remote_name) - 1] = '\0'; /* replacing '\n' */

    printf("\nEnter QoS and Traffic Parameters:\n");
 
 retry_fw_class:
    printf("\nForward Traffic Class: [1=UBR / 2=CBR] [default=UBR]: ");

    if (fgets(choice_str, sizeof(choice_str), stdin) == NULL) {
	diag(COMPONENT, DIAG_ERROR, "I/O ERROR");
	abort();		/* something is going wrong */
    }

    if ((len = strlen(choice_str)) < 2) {
	fw_traffic_class = ATM_UBR;
	goto retry_bw_class;
    }
    choice_str[len - 1] = '\0';	/* replacing '\n' */
    if ((fw_traffic_class = str2num(choice_str)) < 0) {
	diag(COMPONENT, DIAG_ERROR,
	     "input_setup: str2num: %s : Not a number", choice_str);
	goto retry_fw_class;
    }
    if (fw_traffic_class > 2) {
	diag(COMPONENT, DIAG_ERROR,
	     "input_setup: illegal Traffic Class value %s", choice_str);
	goto retry_fw_class;
    }

 retry_bw_class:
    printf("\nBackward Traffic Class: [1=UBR / 2=CBR] [default = UBR]: ");

    if (fgets(choice_str, sizeof(choice_str), stdin) == NULL) {
	diag(COMPONENT, DIAG_ERROR, "I/O ERROR");
	abort();		/* something is going wrong */
    }

    if ((len = strlen(choice_str)) < 2) {
	bw_traffic_class = ATM_UBR;
	goto retry_fw_pcr;
    }
    choice_str[len - 1] = '\0';	/* replacing '\n' */
    if ((bw_traffic_class = str2num(choice_str)) < 0) {
	diag(COMPONENT, DIAG_ERROR,
	     "input_setup: str2num: %s : Not a number", choice_str);
	goto retry_bw_class;
    }
    if (bw_traffic_class > 2) {
	diag(COMPONENT, DIAG_ERROR,
	     "input_setup: illegal Traffic Class value %s", choice_str);
	goto retry_bw_class;
    }
    
 retry_fw_pcr:
    printf("\nMaximum Forward PCR: [default = ATM_OC3_PCR/10 = 35320]: ");
    if (fgets(choice_str, sizeof(choice_str), stdin) == NULL) {
	diag(COMPONENT, DIAG_ERROR, "I/O ERROR");
	abort();		/* something is going wrong */
    }

    if ((len = strlen(choice_str)) < 2) {
	max_fw_pcr = ATM_OC3_PCR/10;		/* cf RFC 1755, par. 7.1 */
	goto retry_bw_pcr;
    }
    choice_str[len - 1] = '\0';	/* replacing '\n' */
    if ((max_fw_pcr = str2num(choice_str)) < 0) {
	diag(COMPONENT, DIAG_ERROR,
	     "input_setup: str2num: %s : Not a number", choice_str);
	goto retry_fw_pcr;
    }
    if (max_fw_pcr > ATM_OC3_PCR) {
	diag(COMPONENT, DIAG_ERROR,
	     "input_setup: illegal PCR value %s", choice_str);
	goto retry_fw_pcr;
    }

 retry_bw_pcr:
    printf("\nMaximum Backward PCR: [default = 35320] : ");
    if (fgets(choice_str, sizeof(choice_str), stdin) == NULL) {
	diag(COMPONENT, DIAG_ERROR, "I/O ERROR");
	abort();		/* something is going wrong */
    }

    if ((len = strlen(choice_str)) < 2) {
	max_bw_pcr = ATM_OC3_PCR/10;
	goto retry_fw_sdu;
    }
    choice_str[len - 1] = '\0';	/* replacing '\n' */
    if ((max_bw_pcr = str2num(choice_str)) < 0) {
	diag(COMPONENT, DIAG_ERROR,
	     "input_setup: str2num: %s : Not a number", choice_str);
	goto retry_bw_pcr;
    }
    if (max_bw_pcr > ATM_OC3_PCR) {
	diag(COMPONENT, DIAG_ERROR,
	 "input_setup: illegal PCR value %s", choice_str);
	goto retry_bw_pcr;
    }
 
 retry_fw_sdu:
    printf("\nForward SDU size (in bytes): [default = 9188]: ");
    if (fgets(choice_str, sizeof(choice_str), stdin) == NULL) {
	diag(COMPONENT, DIAG_ERROR, "I/O ERROR");
	abort();		/* something is going wrong */
    }

    if ((len = strlen(choice_str)) < 2) {
	max_fw_sdu = DEFAULT_MAX_SDU;
	goto retry_bw_sdu;
    }
    choice_str[len - 1] = '\0';	/* replacing '\n' */
    if ((max_fw_sdu = str2num(choice_str)) < 0) {
	diag(COMPONENT, DIAG_ERROR,
	     "input_setup: str2num: %s : Not a number", choice_str);
	goto retry_fw_sdu;
    }
    if (max_fw_sdu > MAX_SDU_SIZE) {
	diag(COMPONENT, DIAG_ERROR, 
	     "input_setup: too big SDU size value: %s > %d", 
	       choice_str, MAX_SDU_SIZE);
	goto retry_fw_sdu;
    }

 retry_bw_sdu:
    printf("\nBackward SDU size (in bytes): [default = 9188]: ");
    if (fgets(choice_str, sizeof(choice_str), stdin) == NULL) {
	diag(COMPONENT, DIAG_ERROR, "I/O ERROR");
	abort();		/* something is going wrong */
    }

    if ((len = strlen(choice_str)) < 2) {
	max_bw_sdu = DEFAULT_MAX_SDU;
	goto fill_in;
    }
    choice_str[len - 1] = '\0';	/* replacing '\n' */
    if ((max_bw_sdu = str2num(choice_str)) < 0) {
	diag(COMPONENT, DIAG_ERROR, 
	     "input_setup: str2num: %s : Not a number", choice_str);
	goto fill_in;
    }
    if (max_bw_sdu > MAX_SDU_SIZE) {
	diag(COMPONENT, DIAG_ERROR, 
	     "input_setup: too big SDU size value: %s > %d", 
	       choice_str, MAX_SDU_SIZE);
	goto retry_bw_sdu;
    }
 
 fill_in:
    /* fill in child_pmsg structure and write command to pipe */
    bzero(&wmsg, sizeof(wmsg));
    wmsg.type = type;
    bcopy(remote_name, wmsg.rname, strlen(remote_name));
    wmsg.fw_traffic_class = (u_int8_t) fw_traffic_class;
    wmsg.bw_traffic_class = (u_int8_t) bw_traffic_class;
    wmsg.max_fw_pcr = max_fw_pcr;
    wmsg.max_bw_pcr = max_bw_pcr;
    wmsg.max_fw_sdu = max_fw_sdu;
    wmsg.max_bw_sdu = max_bw_sdu;

    if ((ret = write(psd[1], &wmsg, sizeof(wmsg))) < 0) {
	diag(COMPONENT, DIAG_ERROR, 
	     "input_setup write: %s", strerror(errno));
	return(ret);
    }
    if (ret != sizeof(wmsg)) {
	diag(COMPONENT, DIAG_ERROR, "input_setup : bad write");
	return(ret);
    }
    return(0);
}

int user_input(void)
{
    int ret = 0, len, choice;
    char choice_str[BUF_LEN], vcc_str[BUF_LEN];
    u_int8_t type, ret_code;
    u_long vcc;
    struct child_pmsg wmsg;

 menu:
    menu_display();
    if (fgets(choice_str, sizeof(choice_str), stdin) == NULL) {
	diag(COMPONENT, DIAG_ERROR, "I/O ERROR on MENU");
	abort();		/* something is going wrong */
    }
    if ((len = strlen(choice_str)) < 2) 
	goto menu;
    choice_str[len - 1] = '\0';	/* replacing '\n' */
    if ((ret = str2num(choice_str)) < 0) {
	diag(COMPONENT, DIAG_ERROR, 
	     "user_input: str2num: %s : Not a number", choice_str);
	fflush(stdout);
	(void) sleep(2);
	goto menu;
    }
    choice = ret;
    type = (u_int8_t) choice;
    switch (choice) {
    case ch_show:		/* (1) show_atmcalls() */
	if ((ret = write(psd[1], &type, 1)) < 0) { /* send only the type */
	    diag(COMPONENT, DIAG_ERROR,
		 "user_input write: %s", strerror(errno));
	    goto menu;
	}
	if (ret != 1) {
	    diag(COMPONENT, DIAG_ERROR, 
		 "user_input (case %d): bad write", type);
	    goto menu;
	}
	if ((ret = read(psd[1], &ret_code, 1)) != 1) {
	    diag(COMPONENT, DIAG_ERROR, 
		 "user_input: read error %s", strerror(errno));
	    goto menu;
	}
	if (ret_code != 0)
	    diag(COMPONENT, DIAG_WARN,
		 "user_input: got trouble when processing command (%d)", 
		   type);
	goto menu;
    case ch_setup:		/* (2) atmcall_setup(ipaddr, conf) */
	if (input_setup() != 0)
	    goto menu;
	if ((ret = read(psd[1], &ret_code, 1)) != 1) {
	    diag(COMPONENT, DIAG_ERROR, 
		 "input_setup: read error %s", strerror(errno));
	    goto menu;
	}
	if (ret_code != 0)
	    diag(COMPONENT, DIAG_WARN,
		 "user_input: got trouble when processing command (%d)", 
		   type);
	goto menu;
    case ch_close:		/* (3) active_close_atmcall(..., vcc, ...) */
    retry_vcc:
	printf("\n vcc : ");
	if (fgets(vcc_str, sizeof(vcc_str), stdin) == NULL) {
	    diag(COMPONENT, DIAG_ERROR, "I/O ERROR");
	    abort();		/* something is going wrong */
	}

	if ((len = strlen(vcc_str)) < 2) 
	    goto retry_vcc;
	vcc_str[len - 1] = '\0'; /* replacing '\n' */
	if ((ret = str2num(vcc_str)) < 0) {
	    diag(COMPONENT, DIAG_ERROR, 
		 "user_input: str2num: %s : Not a number", vcc_str);
	    fflush(stdout);
	    (void) sleep(1);
	    goto menu;
	}
	vcc = (u_long) ret;
	bzero(&wmsg, sizeof(wmsg));
	wmsg.type = type;
	wmsg.vcc = vcc;
	if ((ret = write(psd[1], &wmsg, sizeof(wmsg))) < 0) {
	    diag(COMPONENT, DIAG_ERROR, 
		 "user_input write: %s", strerror(errno));
	    goto menu;
	}
	if (ret != sizeof(wmsg)) {
	    diag(COMPONENT, DIAG_ERROR, 
		 "user_input : bad write");
	    goto menu;
	}
	if ((ret = read(psd[1], &ret_code, 1)) != 1) {
	    diag(COMPONENT, DIAG_ERROR, 
		 "user_input: read error %s", strerror(errno));
	    goto menu;
	}
	if (ret_code != 0)
	    diag(COMPONENT, DIAG_WARN, 
		 "user_input: got trouble when processing command (%d)", type);
	goto menu;
    case ch_close_all:			/* (4) close_all(...) */
	if ((ret = write(psd[1], &type, 1)) < 0) {
	    diag(COMPONENT, DIAG_ERROR, 
		 "user_input write: %s", strerror(errno));
	    goto menu;
	}
	if (ret != 1) {
	    diag(COMPONENT, DIAG_ERROR, 
		 "user_input (case %d): bad write", type);
	    goto menu;
	}
	if ((ret = read(psd[1], &ret_code, 1)) != 1) {
	    diag(COMPONENT, DIAG_ERROR, 
		 "user_input: read error %s", strerror(errno));
	    goto menu;
	}
	if (ret_code != 0)
	    diag(COMPONENT, DIAG_WARN, 
		 "user_input: got trouble when processing command (%d)", type);
	goto menu;
    case ch_quit:			/* (5) QUIT */
	if ((ret = write(psd[1], &type, 1)) < 0) {
	    diag(COMPONENT, DIAG_ERROR, 
		 "user_input write: %s", strerror(errno));
	    goto menu;
	}
	if (ret != 1) {
	    diag(COMPONENT, DIAG_ERROR, 
		 "user_input (case %d): bad write", type);
	    goto menu;
	}
	if ((ret = read(psd[1], &ret_code, 1)) != 1) {
	    diag(COMPONENT, DIAG_ERROR, 
		 "input_setup: read error %s", strerror(errno));
	    goto menu;
	}
	switch(ret_code) {
	case 0:
	    return(1);		/* we are about to quit */
	case 1:
	    goto menu;
	default:
	    diag(COMPONENT, DIAG_WARN, 
		 "user_input: got trouble when processing command (%d)", type);
	    goto menu;
	}
    default:
	diag(COMPONENT, DIAG_ERROR, "user_input: %d BAD CHOICE", choice);
	fflush(stdout);
	(void) sleep(2);
	goto menu;
    }	
}

static int process_p2p_setup(struct child_pmsg *rmsg)
{
    struct in6_addr ria6;
    struct hostent *hp;
    int err_num;
    const char *hstrerr;
    char v[64];
    struct neighbor *ndp;
    struct sockaddr_in6 peer;
    ATMCALL *call;
    struct svc_addr *rsvcaddr;
    struct atm_qos qos;
    int sd, error;

    if (rmsg->rname == NULL) {
	diag(COMPONENT, DIAG_ERROR, "process_p2p_setup: NULL remote name");
	return(-1);
    }
    
    hp = getipnodebyname(rmsg->rname, AF_INET6, AI_DEFAULT, &err_num);
    if (hp == NULL) {
	hstrerr = hstrerror(err_num);
	herror(hstrerr);
	return(-1);
    }
   
    bcopy(hp->h_addr, &ria6, IN6_ADDR_LEN);
    
    diag(COMPONENT, DIAG_DEBUG, 
	 "IPv6 Destination Address: %s", 
	 inet_ntop(AF_INET6, &ria6, v, sizeof(v)));

    ndp = ndplookup(&ria6, 0);
    if (ndp == NULL) {
	if ((sd = socket(PF_INET6, SOCK_DGRAM, 0)) < 0) {
	    diag(COMPONENT, DIAG_ERROR, 
		 "process_p2p_setup: socket (UDP): %s", strerror(errno));
	    return(-1);
	}
	bzero(&peer, sizeof(peer));
	peer.sin6_family = hp->h_addrtype;
        peer.sin6_len = sizeof(peer);
	peer.sin6_port = htons(30000);
	bcopy(hp->h_addr, &peer.sin6_addr, hp->h_length);
	freehostent(hp);

	if (connect(sd, (struct sockaddr *)&peer, sizeof(peer)) < 0) {
	    diag(COMPONENT, DIAG_ERROR,
		 "process_p2p_setup: connect (UDP): %s", strerror(errno));
	    return(-1);
	}

	(void) close(sd);
	
	printf("\n!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!\n");
	printf("process_p2p_setup: JUST COME BACK LATER AND RETRY IT!!!\n");
	printf("!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!\n");
	return(0);
    }

    /* a corresponding neighbor exists */
    /* check for LLINFO existence */
    if ((ndp->flags & ND6F_HAS_LLADDR) == 0) {
	printf("process_p2p_setup: Not ready yet, come back later and retry\n");
	return(0);
    }
    call = (ATMCALL *) ndp->circuit;
    rsvcaddr = &call->rsvcaddr;

    /* fill in qos structure */
    bzero(&qos, sizeof(qos));
    qos.txtp.traffic_class = rmsg->fw_traffic_class;
    qos.txtp.max_pcr = rmsg->max_fw_pcr;
    qos.txtp.max_sdu = rmsg->max_fw_sdu;
    qos.rxtp.traffic_class = rmsg->bw_traffic_class;
    qos.rxtp.max_pcr = rmsg->max_bw_pcr;
    qos.rxtp.max_sdu = rmsg->max_bw_sdu;
    qos.aal = ATM_AAL5;

    error = atmcall_setup(rsvcaddr, rmsg->rname, NULL, &qos,
			  -1, NULL, NULL);
    if (error != 0) {
	diag(COMPONENT, DIAG_ERROR, 
	     "process_p2p_setup: failed to setup ATM connection with peer");
	return(-1);
    }

    return(0);
}

int process_cmd(void)
{
    int ret = 0, len;
    ATMCALL *call;
    u_int8_t type, ret_code;
    struct child_pmsg rmsg;
    pid_t pid;
    int status;
    PARTY *party;

    if ((ret = read(psd[0], &type, 1)) != 1) {
	diag(COMPONENT, DIAG_ERROR, 
	     "process_cmd: read error %s", strerror(errno));
	ret_code = 0;
	if ((ret = write(psd[0], &ret_code, 1)) < 0) {
	    diag(COMPONENT, DIAG_ERROR, 
		 "process_cmd write: %s", strerror(errno));
	    return(-1);
	} 
	return(0);
    }
  
    if ((type == ch_setup) || (type == ch_close)) {
	len = sizeof(rmsg);
	bzero(&rmsg, len);
	if ((ret = read(psd[0], (u_int8_t *)&rmsg + 1, len - 1)) != len - 1) {
	    diag(COMPONENT, DIAG_ERROR, 
		 "process_cmd: read error %s", strerror(errno));
	    ret_code = 0;
	    if ((ret = write(psd[0], &ret_code, 1)) < 0) {
		diag(COMPONENT, DIAG_ERROR, 
		     "process_cmd write: %s", strerror(errno));
		return(-1);
	    } 
	    return(0);
	}
    }

    printf("\n\n");
    switch (type) {
    case ch_show:
	show_atmcalls();
	ret_code = 0;
	break;

    case ch_setup:
	ret = process_p2p_setup(&rmsg);
	ret_code = ret ? 1 : 0;
	break;

    case ch_close:
	ret = active_close_atmcall(0, rmsg.vcc, NULL);
	ret_code = ret ? 1 : 0;
	break;

    case ch_close_all:
	ret = close_all(0);
	ret_code = ret ? 1 : 0;
	break;

    case ch_quit: 
	if (AtmCalls != NULL) {
	    printf("\nprocess_cmd: !!! STILL %d PENDING CALL(S)", NbCalls);
	    printf("\n==> GO BACK TO MENU AND CLOSE ALL CALLS\n");
	    (void) sleep(2);
	    ret_code = 1;
	    break;
	}
	ret_code = 0;
	break;

    default:
	diag(COMPONENT, DIAG_ERROR,
	     "process_cmd: Unknown type of command: %d", type);
	ret_code = 1;
	fflush(stdout);
	(void) sleep(2);
    }
    if ((ret = write(psd[0], &ret_code, 1)) < 0) {
	diag(COMPONENT, DIAG_ERROR, 
	     "process_cmd write: %s", strerror(errno));
	return(-1);
    }
    if ((type == ch_quit) && (ret_code == 0)) {
	printf("\n!!! Terminating session\n");
	if ((pid = wait(&status)) < 0) {
	    diag(COMPONENT, DIAG_ERROR, "wait error: %s", strerror(errno));
	    return(-1);
	}
	return(1);
    }
    return(0);
}
