/* atmcall.h - atm call management */
 
/* Written 1998-1999 by Mohsen.Souissi@inria.fr, INRIA */
 
#ifndef ATMCALL_H
#define ATMCALL_H

#define MAX_CALLS 	10       /* maximum number of ATM calls per client */
#define NAME_LEN 	64
#define DEFAULT_MAX_SDU 9188
#define MAX_SDU_SIZE	65536

typedef enum { 
    null, connecting, accepting, connected, disconnecting
} ATMCALL_STATE;

extern struct atm_qos qos_default_p2p, qos_default_p2m;

typedef struct _call {
  u_long vcc;
  u_int8_t      vpi;
  u_int16_t     vci;
  CONN_CONF 	conf;		/* p2p, root or leaf */
  u_int8_t  	owner;		/* flag = 1, if if call initiator */
  ATMCALL_STATE state;
  int sd;			/* socket desc. for native ATM connection */
#ifndef NDP_CLIENT
  int route;			/* this field has sense only in p2p case */
#else
  struct neighbor *ndp;		/* back pointer to neighbor entry */
#endif
  struct svc_addr rsvcaddr;	/* remote svc_addr if conf != root */
  char rname[NAME_LEN];		/* if p2p/leaf conf and route!=0, remote name */
  struct in6_addr rin6;		/* remote IPv6 address (binary) if provided */
  struct atm_qos qos;		/* corresponding QoS (if any) */

  struct _call *next;
  struct _call *prev;
  int nbactp;
  struct _party *parties;	/* used if root or leaf conf */
#ifdef MARS_CLIENT
  struct resolv_entry *res_e;	/* corresponding resolv_entry structure */
  int revalidate;		/* flag for revalidating the p2mp VC */
  int rt_count;			/* number of times this call has been 
				 * used as a gateway */
#endif
#ifdef MARSD
  struct mcl_entry *mcl_e;	/* corresponding mars client entry */
#endif
} ATMCALL;

typedef enum { ps_null, ps_add_init, ps_add_recv, ps_active,
			   ps_drop_init,  ps_drop_recv } PARTY_STATE;

/* this structure is quite similar to the one in sigd/proto.h */
typedef struct _party {
  u_int16_t ep_ref;		/* endpoint reference must be >= 0 */
  PARTY_STATE pstate;
  struct svc_addr rsvcaddr;
  char rname[NAME_LEN];

  /* --- some point-to-multipoint information ---------------------------- */
  struct _party *next;		/* next party */
  struct _party *prev;		/* previous party */
  struct _call *call;		/* back pointer to the ATMCALL struct */
#ifdef MARS_CLIENT
  struct ept_entry *ept_e;	/* corresponding ept_entry structure */
#endif
#ifdef MARSD
  struct mcl_entry *mcl_e;	/* corresponding mars client entry */
#endif
} PARTY;


extern ATMCALL *AtmCalls;     /* pointer to the last established call */
extern int NbCalls;
extern u_long CurrVcc;
extern const char *CallStateName[];
extern const char *PartyStateName[];

PARTY *new_party(ATMCALL *call, u_int16_t ep_ref, PARTY_STATE pstate, 
		 struct svc_addr *remote_atm, char *remote_name, 
		 void *ept_e, void *clmb_e);
ATMCALL *new_atmcall(u_long vcc, u_int8_t vpi, u_int16_t vci, 
		     CONN_CONF conf, u_int8_t owner, ATMCALL_STATE state, 
		     struct svc_addr *remote_atm, char *remote_name, 
		     struct in6_addr *remote_in6, struct atm_qos *qos,
		     void *clnt_e);

void append_atmcall(ATMCALL *call);
void show_atmcalls(void);
ATMCALL *atmcall_lookup(u_long vcc);
ATMCALL *atmcall_lookup2(int sd);
#ifdef NDP_CLIENT
ATMCALL *atmcall_lookupa(struct svc_addr *addr);
#endif
PARTY *party_lookup(u_long vcc, u_int16_t ep_ref);

void free_party(PARTY *party);
void free_all_parties(ATMCALL *call);
void free_atmcall(ATMCALL *call);

int atmcall_setup(struct svc_addr *remote_atm, char *remote_name, 
		  struct in6_addr *remote_in6, struct atm_qos *qos,
		  int16_t ep_ref, void *ept_e, void *mcl_e);

PARTY *active_add_party(ATMCALL *call, struct svc_addr *remote_atm,
                        char *remote_name, void *ept_e, void * mcl_e);

int new_natm_sock(ATMCALL *call);

int active_drop_party(u_long vcc, u_int16_t ep_ref, PARTY *partyarg);
int active_close_atmcall(int sig_down, u_long vcc, ATMCALL *callarg);
int close_all(int sig_down);

#ifndef NDP_CLIENT
int add_ipatmroute(int is_host, 
		   struct in6_addr *rin6, 
		   int plen, 
		   ATMCALL *call,
		   u_int8_t rtm_op);
#endif
ATMCALL *delete_ipatmroute(u_long vcc, 
			   int is_host, 
			   struct in6_addr *rin6, 
			   int plen, 
			   ATMCALL *callarg);
#ifndef NDP_CLIENT
int sendrtmsg(int op, 
	      int is_host, 
	      struct in6_addr *rin6, 
	      int plen, 
	      u_int8_t vpi,
	      u_int16_t vci);
#endif

#endif

