/* scl_menu.c - Interface MENU for generic ATM signaling client */

/* Written 1998-1999 by Mohsen.Souissi@inria.fr, INRIA */

#include <sys/types.h>
#include <sys/socket.h>
#include <net/if.h>
#include <netinet/in.h>
#include <netnatm/natm.h>
#include <arpa/inet.h>
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <signal.h>
#include <limits.h>
#include <sys/wait.h>

#include "../linux-atm.h"
#include "atmd.h"
#include "atm.h"
#include "misc_tools.h"

#include "ndp.h"
#include "ipatmlookup.h"
#include "atmcall.h"
#include "iponatm.h"
#include "scl_io.h"

#define COMPONENT	"SCL_MENU"

int psd[2];			/* socket pair descriptors */

static void menu_display(void) 
{
    (void) sleep(1);
    printf("\n\n");
    printf("\t\t****************************************\n");
    printf("\t\t*          ATM SIGNALING API           *\n");
    printf("\t\t****************************************\n\n\n");
    printf("1) SHOW Current ATM Calls\n");
    printf("2) SETUP a New ATM Call\n");
    printf("3) ADD an IP Route\n");
    printf("4) DELETE an IP Route\n");
    printf("5) CLOSE an ATM Call (deletes route if any)\n");
    printf("6) CLOSE ALL \n");
    printf("7) ADD a Party to an ATM Call\n");
    printf("8) DROP a Party from an ATM Call \n");
    printf("9) QUIT\n");
    printf("\nchoice: ");
}

static int input_setup(void)		/* Setup a new ATM Call */
{
    int ret;
    int num, choice, len;
    char choice_str[BUF_LEN], remote_name[NAME_LEN];
    CONN_CONF conf;
    u_int8_t type = 2;
    u_int8_t *remote_aesa;
    PIPE_MSG wmsg;

    printf("\n\n\n\n");
    printf("1) Point-to-Point ATM Call\n");
    printf("2) Point-to-Multipoint ATM Call\n");
    printf("3) Back to MENU\n");
    printf("\nchoice: ");
    if (fgets(choice_str, sizeof(choice_str), stdin) == NULL) 
	return(input_setup());
    if ((len = strlen(choice_str)) < 2)
	return(input_setup());
    choice_str[len - 1] = '\0';	/* replacing '\n' */
    if ((num = str2num(choice_str)) < 0) {
	printf("\n input_setup: str2num: %s : not a number", choice_str);
	return(input_setup()); 
    }
    else {
	choice = num;
	switch (choice) {
	case 1:
	    conf = p2p;
	    break;
	case 2:
	    conf = root;
	    break;
	case 3:
	    return(1);		/* will go back to menu_display */
	default:
	    printf("\ninput_setup: %d BAD CHOICE\n", choice);
	    return(input_setup());
	}
    }

    printf("\n Remote host : ");
    if (fgets(remote_name, sizeof(remote_name), stdin) == NULL)
	return(1);

    remote_name[strlen(remote_name) - 1] = '\0'; /* replacing '\n' */
    /* address resolution (local lookup) */
    if ((remote_aesa = getaesabyname(AddrConf, remote_name)) == NULL) {
        printf("\n input_setup: getaesabyname: %s : Unknown host\n", 
               remote_name);
        return(1);
    }

    /* write command to pipe */
    bzero(&wmsg, sizeof(wmsg));
    wmsg.type = type;
    bcopy(remote_name, wmsg.rname, strlen(remote_name));
    bcopy(remote_aesa, wmsg.raesa, ATM_ESA_LEN);
    wmsg.conf = conf;
    if ((ret = write(psd[1], &wmsg, sizeof(wmsg))) < 0) {
	fprintf(stderr, "input_setup write: %s\n", strerror(errno));
	return(ret);
    }
    if (ret != sizeof(wmsg)) {
	printf("ERROR: input_setup : bad write\n");
	return(ret);
    }
    return(0);
}

static int input_add_party(void)      /* case 7 */
{
    int ret, len;
    char vcc_str[BUF_LEN], remote_name[NAME_LEN];
    u_long vcc;
    u_int8_t type = 7;
    u_int8_t *remote_aesa;
    PIPE_MSG wmsg;

    printf("\n\n\n\n");
    printf("\n Remote host : ");
    if (fgets(remote_name, sizeof(remote_name), stdin) == NULL)
	return(1);
  
    remote_name[strlen(remote_name) - 1] = '\0'; /* replacing '\n' */
    if ((remote_aesa = getaesabyname(AddrConf, remote_name)) == NULL) {
        printf("\n input_add_party: getaesabyname: %s : Unknown host\n", 
               remote_name);
        return(1);
    }
    printf("\n vcc : ");
    if (fgets(vcc_str, sizeof(vcc_str), stdin) == NULL) 
	return(2);
    len = strlen(vcc_str);
    vcc_str[len - 1] = '\0';	/* replacing '\n' */
    if ( (len < 2) || ((ret = str2num(vcc_str)) < 0) ) {
	printf("\ninput_add_party: str2num: %s : not a number", vcc_str);
	fflush(stdout);
	(void) sleep(1);
	return(3);
    }
    vcc = (u_long) ret;
 
    /* write command to pipe */
    bzero(&wmsg, sizeof(wmsg));
    wmsg.type = type;
    wmsg.vcc = vcc;
    bcopy(remote_name, wmsg.rname, strlen(remote_name));
    bcopy(remote_aesa, wmsg.raesa, ATM_ESA_LEN);
    if ((ret = write(psd[1], &wmsg, sizeof(wmsg))) < 0) {
	fprintf(stderr, "input_add_party write: %s\n", strerror(errno));
	return(4);
    }
    if (ret != sizeof(wmsg)) {
	printf("ERROR: input_add_party : bad write\n");
	return(5);
    }
    return(0);
}

static int input_drop_party(void)		/* case 8 */
{
    int ret, len;
    char vcc_str[BUF_LEN], epr_str[BUF_LEN];
    u_long vcc;
    u_short ep_ref;
    PIPE_MSG wmsg;
    u_int8_t type = 8;

    printf("\n\n\n\n");
  
    printf("\n vcc : ");
    if (fgets(vcc_str, sizeof(vcc_str), stdin) == NULL) 
	return(1);
    len = strlen(vcc_str);
    vcc_str[len - 1] = '\0';	/* replacing '\n' */
    if ( (len < 2) || ((ret = str2num(vcc_str)) < 0) ) {
	printf("\ninput_drop_party: str2num: %s : not a number", vcc_str);
	fflush(stdout);
	(void) sleep(1);
	return(2);
    }
    vcc = (u_long) ret;

    printf("\n ep_ref : ");
    if (fgets(epr_str, sizeof(epr_str), stdin) == NULL) 
	return(3);
    len = strlen(epr_str);
    epr_str[len - 1] = '\0';	/* replacing '\n' */
    if ( (len < 2) || ((ret = str2num(epr_str)) < 0) ) {
	printf("\ninput_drop_party: str2num: %s : not a number", epr_str);
	fflush(stdout);
	(void) sleep(1);
	return(4);
    }
    ep_ref = (u_short) ret;
 
    /* write command to pipe */
    bzero(&wmsg, sizeof(wmsg));
    wmsg.type = type;
    wmsg.vcc = vcc;
    wmsg.ep_ref = ep_ref;
    if ((ret = write(psd[1], &wmsg, sizeof(wmsg))) < 0) {
	fprintf(stderr, "input_drop_party write: %s\n", strerror(errno));
	return(5);
    }
    if (ret != sizeof(wmsg)) {
	printf("ERROR: input_drop_party : bad write\n");
	return(6);
    }
    return(0);
}

int user_input(void)
{
    int ret = 0, len, choice;
    char choice_str[BUF_LEN], vcc_str[BUF_LEN];
    u_int8_t type, ret_code;
    u_long vcc;
    PIPE_MSG wmsg;

 menu:
    menu_display();
    if (fgets(choice_str, sizeof(choice_str), stdin) == NULL) {
	printf("I/O ERROR on MENU\n");
	/* goto menu; */
	abort();		/* something is going wrong */
    }
    if ((len = strlen(choice_str)) < 2) 
	goto menu;
    choice_str[len - 1] = '\0';	/* replacing '\n' */
    if ((ret = str2num(choice_str)) < 0) {
	printf("\nuser_input: str2num: %s : Not a number\n", choice_str);
	fflush(stdout);
	(void) sleep(2);
	goto menu;
    }
    choice = ret;
    type = (u_int8_t) choice;
    switch (choice) {
    case 1:			/* show_atmcalls() */
	if ((ret = write(psd[1], &type, 1)) < 0) { /* send only the type */
	    fprintf(stderr, "user_input write: %s\n", strerror(errno));
	    goto menu;
	}
	if (ret != 1) {
	    printf("ERROR: user_input (case %d): bad write\n", type);
	    goto menu;
	}
	if ((ret = read(psd[1], &ret_code, 1)) != 1) {
	    printf("ERROR:  user_input: read error %s\n", strerror(errno));
	    goto menu;
	}
	if (ret_code != 0)
	    printf("user_input: got trouble when processing command (%d)\n", 
		   type);
	goto menu;
    case 2:			/* atmcall_setup(ipaddr, conf) */
	if (input_setup() != 0)
	    goto menu;
	if ((ret = read(psd[1], &ret_code, 1)) != 1) {
	    printf("ERROR:  input_setup: read error %s\n", strerror(errno));
	    goto menu;
	}
	if (ret_code != 0)
	    printf("user_input: got trouble when processing command (%d)\n", 
		   type);
	goto menu;
    case 3:			/* ipoveratm(..., vcc, ...) */
    case 4:			/* delete_ipatmroute(vcc, ...) */
    case 5:			/* active_close_atmcall(..., vcc, ...) */
	printf("\n vcc : ");
	if (fgets(vcc_str, sizeof(vcc_str), stdin) == NULL) 
	    goto menu;
	len = strlen(vcc_str);
	vcc_str[len - 1] = '\0'; /* replacing '\n' */
	if ( (len < 2) || ((ret = str2num(vcc_str)) < 0) ) {
	    printf("\nuser_input: str2num: %s : not a number", vcc_str);
	    fflush(stdout);
	    (void) sleep(1);
	    goto menu;
	}
	vcc = (u_long) ret;
	bzero(&wmsg, sizeof(wmsg));
	wmsg.type = type;
	wmsg.vcc = vcc;
	if ((ret = write(psd[1], &wmsg, sizeof(wmsg))) < 0) {
	    fprintf(stderr, "user_input write: %s\n", strerror(errno));
	    goto menu;
	}
	if (ret != sizeof(wmsg)) {
	    printf("ERROR: user_input : bad write\n");
	    goto menu;
	}
	if ((ret = read(psd[1], &ret_code, 1)) != 1) {
	    printf("ERROR:  user_input: read error %s\n", strerror(errno));
	    goto menu;
	}
	if (ret_code != 0)
	    printf("user_input: got trouble when processing command (%d)\n", 
		   type);
	goto menu;
    case 6:			/* close_all(...) */
	if ((ret = write(psd[1], &type, 1)) < 0) {
	    fprintf(stderr, "user_input write: %s\n", strerror(errno));
	    goto menu;
	}
	if (ret != 1) {
	    printf("ERROR: user_input (case %d): bad write\n", type);
	    goto menu;
	}
	if ((ret = read(psd[1], &ret_code, 1)) != 1) {
	    printf("ERROR:  user_input: read error %s\n", strerror(errno));
	    goto menu;
	}
	if (ret_code != 0)
	    printf("user_input: got trouble when processing command (%d)\n", 
		   type);
	goto menu;
    case 7:
	if (input_add_party() != 0)
	    goto menu;
	if ((ret = read(psd[1], &ret_code, 1)) != 1) {
	    printf("ERROR:  process_cmd (case %d): read error %s\n",type,  
		   strerror(errno));
	    goto menu;
	}
	if (ret_code != 0)
	    printf("user_input: got trouble when processing command (%d)\n", 
		   type);
	goto menu;
    case 8:
	if (input_drop_party() != 0)
	    goto menu;
	if ((ret = read(psd[1], &ret_code, 1)) != 1) {
	    printf("ERROR:  process_cmd (case %d): read error %s\n",type,  
		   strerror(errno));
	    goto menu;
	}
	if (ret_code != 0)
	    printf("user_input: got trouble when processing command (%d)\n", 
		   type);
	goto menu;
    case 9:			/* QUIT */
	if ((ret = write(psd[1], &type, 1)) < 0) {
	    fprintf(stderr, "user_input write: %s\n", strerror(errno));
	    goto menu;
	}
	if (ret != 1) {
	    printf("ERROR: user_input (case %d): bad write\n", type);
	    goto menu;
	}
	if ((ret = read(psd[1], &ret_code, 1)) != 1) {
	    printf("ERROR:  input_setup: read error %s\n", strerror(errno));
	    goto menu;
	}
	switch(ret_code) {
	case 0:
	    return(1);		/* we are about to quit */
	case 1:
	    goto menu;
	default:
	    printf("user_input: got trouble when processing command (%d)\n", 
		   type);
	    goto menu;
	}
    default:
	printf("\nuser_input: %d BAD CHOICE\n", choice);
	fflush(stdout);
	(void) sleep(2);
	goto menu;
    }	
}

int process_cmd(void)
{
    int ret = 0, len;
    ATMCALL *call;
    u_int8_t type, ret_code;
    PIPE_MSG rmsg;
    pid_t pid;
    int status;
    PARTY *party;
    struct svc_addr remote_atm;

    if ((ret = read(psd[0], &type, 1)) != 1) {
	printf("ERROR:  process_cmd: read error %s\n", strerror(errno));
	ret_code = 0;
	if ((ret = write(psd[0], &ret_code, 1)) < 0) {
	    fprintf(stderr, "process_cmd write: %s\n", strerror(errno));
	    return(-1);
	} 
	return(0);
    }
  
    if ((type != 1) && (type != 6) && (type != 9)) {
	len = sizeof(rmsg);
	bzero(&rmsg, len);
	if ((ret = read(psd[0], (u_int8_t *)&rmsg + 1, len - 1)) != len - 1) {
	    printf("ERROR:  process_cmd: read error %s\n", strerror(errno));
	    ret_code = 0;
	    if ((ret = write(psd[0], &ret_code, 1)) < 0) {
		fprintf(stderr, "process_cmd write: %s\n", strerror(errno));
		return(-1);
	    } 
	    return(0);
	}
    }

    printf("\n\n");
    switch (type) {
    case 1:
	show_atmcalls();
	ret_code = 0;
	break;
    case 2:
	bzero(&remote_atm, sizeof(struct svc_addr));
	bcopy(rmsg.raesa, remote_atm.prv, ATM_ESA_LEN);
	ret = atmcall_setup(&remote_atm, rmsg.rname, NULL,
			    (rmsg.conf == root) ? &qos_default_p2m 
			    : &qos_default_p2p,
			    (rmsg.conf == root) ? 0 : -1, NULL, NULL);
	ret_code = ret ? 1 : 0;
	break;
    case 3:
	ret = ipoveratm(NDP_ACTIVE, rmsg.vcc, ndp_dump);
	ret_code = ret ? 1 : 0;
	break;
    case 4:
	ret = (call = delete_ipatmroute(rmsg.vcc, 1, NULL, 0, NULL)) ? 0 : -1;
	ret_code = ret ? 1 : 0;
	break;
    case 5:
	ret = active_close_atmcall(0, rmsg.vcc, NULL);
	ret_code = ret ? 1 : 0;
	break;
    case 6:
	ret = close_all(0);
	ret_code = ret ? 1 : 0;
	break;
    case 7:
	if ((call = atmcall_lookup(rmsg.vcc)) == NULL) {
	    printf("process_cmd: no matching atm call for VCC %ld\n", 
		   rmsg.vcc);
	    ret_code = 1;
	    break;
	}
	if (call->conf != root) {
	    printf("process_cmd: ATM call is not in p2m root conf\n");
	    ret_code = 2;
	    break;
	}
	bzero(&remote_atm, sizeof(struct svc_addr));
	bcopy(rmsg.raesa, remote_atm.prv, ATM_ESA_LEN);
	party = active_add_party(call, &remote_atm, rmsg.rname, NULL, NULL);
	ret_code = party ? 0 : 3;
	break;
    case 8:
	ret = active_drop_party(rmsg.vcc, rmsg.ep_ref, NULL);
	ret_code = ret ? 1 : 0;
	break;
    case 9: 
	if (AtmCalls != NULL) {
	    printf("\nprocess_cmd: !!! STILL %d PENDING CALL(S)", NbCalls);
	    printf("\n==> GO BACK TO MENU AND CLOSE ALL CALLS\n");
	    (void) sleep(2);
	    ret_code = 1;
	    break;
	}
	ret_code = 0;
	break;
    default:
	printf("process_cmd: Unknown type of command: %d\n", type);
	ret_code = 1;
	fflush(stdout);
	(void) sleep(2);
    }
    if ((ret = write(psd[0], &ret_code, 1)) < 0) {
	fprintf(stderr, "process_cmd write: %s\n", strerror(errno));
	return(-1);
    }
    if ((type == 9) && (ret_code == 0)) {
	printf("\n!!! Terminating session\n");
	if ((pid = wait(&status)) < 0) {
	    fprintf(stderr, "wait error: %s\n", strerror(errno));
	    return(-1);
	}
	return(1);
    }
    return(0);
}

